'''
Canonical street names!
(To allow name matching between OSM and Portland City DB.)
'''
import string

import pylev

# ratio of edit distance between strings to string length (= max number of edits possible)
def levenshtein_ratio(s1, s2):
    assert len(s1) > 0 and len(s2) > 0
    return pylev.levenshtein(s1, s2) / float(max(len(s1),len(s2)))

# guess at probability names refer to the same street
def match_probability(names1, names2):
    assert len(names1) > 0 and len(names2) > 0
    p = 0

    for name1 in names1:
        can1 = canonical_street_name(name1)
        for name2 in names2:
            can2 = canonical_street_name(name2)
            #print 'DEBUG', name1, name2, can1, can2, pylev.levenshtein(can1,can2)
            #print 'DEBUG ratio:', levenshtein_ratio(can1,can2)
            p = max(p, 1-levenshtein_ratio(can1,can2))

    return p

# rate name match on scale of 0 to 100
def match_score(names1,names2):
    return round(100*match_probability(names1,names2))

# TODO: consider adding USPS canonical abbreviations for street types
# http://cpansearch.perl.org/src/SDERLE/Geo-Coder-US-0.21/US/Codes.pm
# cped to portland_peds/Codes.pm.txt 
rewrites = {
		'NORTH':'N',
		'SOUTH':'S',
		'EAST':'E',
		'WEST':'W',
		'NORTHEAST':'NE',
		'NORTHWEST':'NW',
		'SOUTHEAST':'SE',
		'SOUTHWEST':'SW',
	
		'AVENUE':'AVE',
                'BRIDGE':'BRG',
		'BOULEVARD':'BLVD',
		'CIRCLE':'CIR',
		'COURT':'CT',
		'DRIVE':'DR',
                'FREEWAY':'',
                'FWY':'',
		'HIGHWAY':'HWY',
                'JUNIOR':'JR',
		'LANE':'LN',
		'MOUNT':'MT',
		'PARKWAY':'PKWY',
		'PLACE':'PL',
		'ROAD':'RD',
		'SAINT':'ST',
		'STREET':'ST',
                'SUMMIT':'SMT',
		'TERRACE':'TER',
                'TRAIL':'TRL'
 		}

def restrict_chars(name):
    ALLOWED_CHARS = string.ascii_letters + string.digits + " "  # others get mapped to ' '
    IGNORE_CHARS = "'"
    out = []
    for c in name:
        if c in ALLOWED_CHARS: 
            out.append(c) 
        else:
            if c not in IGNORE_CHARS: out.append(' ')
    return ''.join(out)

def canonical_street_name(name):
    '''
    Regularizes street name to facilitate comparison.  
    Converts to all caps and applies standard abbreviations.
    '''
    if name is None: return None;

    #print "DEBUG name in: ", name
    space = ' ' 

    # restrict character set: simplicity, security
    name = restrict_chars(name)

    # all caps
    name = name.upper()

    # words
    words=name.strip().split(space)
    words = [w for w in words if w != '']
    #print "DEBUG words: ", words

    # rewrites
    new_words = []
    for word in words:
            if rewrites.has_key(word):
                    new=rewrites[word]
                    if len(new)>0: new_words.append(new)
                    continue
            if len(word)>1 and word[0]=='I' and word[1:].isdigit(): 
                    new_words.append('I')
                    new_words.append(word[1:])
                    continue
            new_words.append(word)
    words = new_words

    # two word rewrites
    new_words = []
    skip=False
    for i in range(len(words)):
        if skip:
            skip=False
            continue
        if i == len(words)-1:
            #last word
            new_words.append(words[i])
            continue
        if words[i] == 'TRANSIT' and words[i+1] == 'CENTER':
            new_words.append('TC')
            skip=True
            continue
        if words[i] == 'UNITED' and words[i+1] == 'STATES':
            new_words.append('US')
            skip=True
            continue
        if words[i+1].isdigit() and words[i] in ['HWY','US','OR']:
            new_words.append('HWY')
            new_words.append(words[i+1])
            skip=True
            continue
        if words[i] == 'MC':
            new_words.append('MC' + words[i+1])
            skip=True
            continue
        if words[i] == 'LE':
            new_words.append('LE' + words[i+1])
            skip=True
            continue
        new_words.append(words[i])
    words = new_words

    # single spaces between words
    out = ' '.join(words).strip()

    return out

def test_can(name1,name2=None):
    can1 = canonical_street_name(name1)
    #print "test_can, name1: '%s'  can1: '%s'" % (name1, can1) 
    assert can1 == canonical_street_name(can1)

    if name2:	
	can2 = canonical_street_name(name2)
	#print "test_can, name2: '%s'  can2: '%s'" % (name2, can2) 
	assert can2 == canonical_street_name(can2)
	assert can1 == can2
		
def test():
	test_can("Southeast Sunburst Lane","SE SUNBURST LN")
	test_can("Northeast Saint George Street","NE ST GEORGE ST")
	test_can("NE HALSEY ST FRONTAGE RD","Northeast Halsey Street Frontage Road")
        test_can("I84 FWY","I 84")
        test_can(" 13133","13133")
        test_can("US BANCORP COLUMBIA CENTER TC","UNITED STATES BANCORP COLUMBIA CENTER TRANSIT CENTER")
        test_can("COEUR DALENE DR", "COEUR D'ALENE DR")
        test_can("BURNSIDE BRG", "BURNSIDE BRIDGE")
        test_can("BLAZER TRL","BLAZER TRAIL")
        test_can("HWY 30","US 30")
        test_can("MARTIN LUTHER KING  JUNIOR BLVD","MARTIN LUTHER KING JR BLVD")
        test_can("Mc Gee Street", "McGee Street")
        test_can("Saint Alban's Road", "ST ALBANS RD")
        test_can("Le Roy Avenue", "LeRoy Avenue")

        # 'fuzzy matching' (edit distance)
        assert match_score(['A St'],['B Street']) == 75.0


	print 'cannames PASS'
#doit
test()



#!/usr/bin/env python

'''
Convert shapefile to geojson
'''

import sys
import itertools
import json 

import shapefile #pip pyshp

import geofeatures

def shp2geojson(inFilename,outFilename,clip_rect=None):
    # read the shapefile
    reader = shapefile.Reader(inFilename)
    fields = reader.fields[1:]
    fieldNames = [field[0] for field in fields]
    #print 'fieldNames', fieldNames
    features = []
    for (sr, ss) in itertools.izip(reader.iterRecords(), reader.iterShapes()):
        atr = dict(zip(fieldNames, sr))
        geom = ss.__geo_interface__
        if not clip_rect or geofeatures.coordinates_intersect_rect_q(geom['coordinates'],clip_rect):
            #print 'DEB geom:', geom
            features.append(dict(type='Feature', geometry=geom, properties=atr)) 
 
    # write the geojson file
    jsonFile = open(outFilename, 'w')
    jsonFile.write(
            json.dumps({'type':'FeatureCollection', 'features':features}, indent=2) 
            + "\n")
    jsonFile.close()

def test():
    print 'shp2geojson PASS'

if __name__ == '__main__':
    if len(sys.argv) == 3:
        shp2geojson(sys.argv[1],sys.argv[2])
    else:
        print >> sys.stderr, 'Usage: %s foo.shp foo.geojson' % sys.argv[0]



''' 
spiderosm config settings
'''

import os
import json

settings = {}

# location of spiderosm package dir
# assume current directory if 'spiderosm' module not loaded
try:
    import spiderosm
    settings['spiderosm_dir'] = spiderosm.__path__[0]
except ImportError:
    settings['spiderosm_dir'] = os.getcwd()

# data for spiderosm tests
settings['spiderosm_test_data_dir'] = os.path.join(settings['spiderosm_dir'],'test_data')

#enabling postgis or spatialite causes results to be written to these database by toplevels 
# such as examples/match_berkeley.py and examples/match_portland.py.  
# examples/test_spiderosm also will test enabled database interfaces
settings['postgis_enabled'] = False
settings['spatialite_enabled'] = False

# main programs call this explictly after setting 'default' conf values.
def read_config_files(filenames=None,paths=None,quiet=False):
    if not filenames: filenames = ('.spiderosm.json', 'config.spiderosm.json')
    if not paths: paths = ('~', '.')

    for path in paths:
        path = os.path.expanduser(path)
        for fn in filenames:
            fullpath = os.path.join(path,fn)
            if os.path.exists(fullpath):
                print 'Reading %s' % fullpath
                with open(fullpath, 'r') as f:
                    new = json.load(f)
                    settings.update(new)

def test():
    fns = ( '.spiderosm.json', 'config.spiderosm.json' )
    paths = ( '~','.' )
    #read_config_files(fns,paths)
    #print 'settings', settings
    print 'config PASS'

#doit
test()






'''
Path Network - Match Scoring
'''

import geo 
import cannames
import pnwk_network

def _prob_to_score(p):
    assert 0 <= p <= 1
    return round(100*p)

class PNwkScore(pnwk_network.PNwkNetwork):
    def score_matches(self):
        for seg in self.segs.values(): 
            seg.score_match()

    class Seg(pnwk_network.PNwkNetwork.Seg):
        def _score_name_match(self):
            if not self.match: return
            if len(self.names)==0 or len(self.match.names)==0: return
            self._set_score('score_name', cannames.match_score(self.names,self.match.names))

        def _score_geo_match(self):
            if not self.match: return
            f = self.network.units_per_meter
            delta = 3.0*f # distance between samples
            if self.match_rev:
                (div,avg_bearing_delta) = geo.divergence(self.points,self.match.points[::-1],delta)
            else:
                (div, avg_bearing_delta) = geo.divergence(self.points,self.match.points,delta)

            # p1 based on divergence  (div of 30 meters or more scores 0)
            p1 = 1.0 - min(div*f,30.0)/30.0
            #print 'p1,div', p1, div

            # p2 based on ratio of divergenece to segment length
            length = min(self.length(),self.match.length())
            if(length<self.network.stepScale): 
                p2 = 0.0
            else:
                p2 = max(0,1.0-div/length)

            score = round(p1*p2*100.0)
            #print 'seg_id,length,div,p1,p2,geoScore',self.seg_id,length,div,p1,p2,score
            self._set_score('divergence', div)
            self._set_score('avg_bearing_delta', avg_bearing_delta)
            self._set_score('score_bearing1', _prob_to_score(1.0 - avg_bearing_delta/180.0))
            self._set_score('score_geo1', _prob_to_score(p1))
            self._set_score('score_geo2', _prob_to_score(p2))

        def _score_bearing_match(self):
            if not self.match: return
            if self.match_rev:
                delta1 = geo.compass_bearing_delta(self.from_bearing,self.match.to_bearing)
                delta2 = geo.compass_bearing_delta(self.to_bearing,self.match.from_bearing)
            else:
                delta1 = geo.compass_bearing_delta(self.from_bearing,self.match.from_bearing)
                delta2 = geo.compass_bearing_delta(self.to_bearing,self.match.to_bearing)

            p1 = 1 - (delta1/180.0)
            p2 = 1 - (delta2/180.0)
            self._set_score('score_bearing2', _prob_to_score(p1*p2))


        def _set_score(self,key,value):
            ns = self.network.match_namespace
            self.tags[ns+key] = value

        def _score_to_factor(self,key,importance=1.0):
            ns = self.network.match_namespace
            key =  ns + key
            if key in self.tags:
                score = self.tags[key]
            else:
                score = 50
            f = float(score)/100.0
            fscaled = (1.0-importance) + importance*f 
            return fscaled

        def score_combine(self,
                score_name_i=0.20,
                score_geo1_i=0.50,
                score_geo2_i=0.50,
                score_bearing1_i=0.50,
                score_bearing2_i=0.50):
            p = (self._score_to_factor('score_name',importance=score_name_i) *
                    self._score_to_factor('score_geo1',importance=score_geo1_i) *
                    self._score_to_factor('score_geo2',importance=score_geo2_i) *
                    self._score_to_factor('score_bearing1',importance=score_bearing1_i) *
                    self._score_to_factor('score_bearing2',importance=score_bearing2_i)
                    )
            return _prob_to_score(p)

        def score_match(self, name_importance=0.20):
            if not self.match: return
            self._score_name_match()
            self._score_geo_match()
            self._score_bearing_match()
            self._set_score('score', self.score_combine())

def test():
    
    # _prob_to_score()
    assert _prob_to_score(0)==0 and _prob_to_score(1)==100 and _prob_to_score(.25)==25 and _prob_to_score(.007)==1

    # test _score_to_factor
    class SegStub(PNwkScore.Seg):
        def __init__(self):
            self.tags={}
    segy = SegStub()
    segy.network = PNwkScore(name='t')
    segy._set_score('0',0)
    segy._set_score('50',50)
    segy._set_score('75',75)
    segy._set_score('80',80)
    segy._set_score('100',100)
    assert segy.tags['match$50']==50
    assert segy._score_to_factor('75') == .75
    assert segy._score_to_factor('not_there') == .50
    assert segy._score_to_factor('0',importance=.1) == .90
    assert segy._score_to_factor('100',importance=.1) == 1.00 
    assert segy._score_to_factor('80',importance=.1) == .98
    assert segy._score_to_factor('not_there',importance=.80) == .60

    print 'pnwk_score PASS'

#doit
test()




import geojson

import colspecs
import geo


class DatabaseInterface(object):
    ''' Base class for postgis, and spatiallite database interfaces'''

    def __init__(self, db_name, srid=None, verbose=True):
        self.db_name = db_name
        if not srid:  srid = 4326 # default to WGS84 longlat
        self.srid = srid
        self.verbose = verbose
        if self.verbose: print 'Connecting to database %s' % db_name
        (self.con, self.cur) = self._connect()
        self._add_spatial_extension()


    def _connect(self):
        print 'DatabaseInterface._connect() stub: specialize for postgis, spatialite etc.'

    def _add_spatial_extension(self):
        print 'DatabaseInterface._add_spatial_extension() stub: specialize for postgis, spatialite etc.'

    # '%s' in pycopg2, '?' in pyspatialite
    _interpolation_ref = 'string used to reference values to be interpolated into sql string by cursor' 

    def code_point(self,point,srid=None):
        if not srid: srid=self.srid
	return "ST_GeomFromText('%s',%d)" % (geo.wkt_point(point), srid)

    def code_linestring(self,points,srid=None):
        if not srid: srid=self.srid
	return "ST_GeomFromText('%s',%d)" % (geo.wkt_linestring(points), srid)

    # NOTE: transactions are initiated automatically (by default)
    def commit(self):
	self.con.commit()

    def close(self):
	self.commit()
	self.con.close()

    # values are interpolated into s, similar to "format % values", but quoted/escaped properly for sql
    # important to use this rather than '%' or '+' for security:  sql injection attacks!
    def exec_sql(self,s,values=None):
	if values:
		self.cur.execute(s,values)
	else:
		self.cur.execute(s)

    # exec sql and return results
    def exec_sql_r(self,s,values=None):
	'''exec sql and return resulting table'''
	self.exec_sql(s, values)
	return self.cur.fetchall()

    def select(self,cols,table,where=None):
	''' make common sql 'SELECT' a tad less cumbersome'''
	parms = {'cols':cols, 'table':table, 'where':where}
	if where:
		return self.exec_sql_r("SELECT %(cols)s FROM %(table)s WHERE %(where)s" % parms)
	else:
		return self.exec_sql_r("SELECT %(cols)s FROM %(table)s" % parms)

    def get_table(self,table):
	rows = self.select('*',table);
	return rows

    def get_table_names(self):
        rows = self.select('table_name','information_schema.tables')
        return [ row[0] for row in rows ]
    
    def get_num_rows(self,table):
	rows = self.exec_sql_r('select count(*) as count from %s' % table)
	assert len(rows)==1 
        #print 'DEBUG rows', rows
	return rows[0]['count']

    def add_column(self,table,name,type='text'):
        self.exec_sql('ALTER TABLE %(table)s ADD COLUMN "%(name)s" %(type)s;' % 
                {'table':table, 'name':name, 'type':type})

    def add_geometry_column(self,table,name,type,srid=None):
        if not srid: srid = self.srid
        #print 'DEB add_geometry_column',table,name,type
        self.exec_sql('''
                SELECT AddGeometryColumn('%(table)s', '%(name)s', %(srid)d, '%(type)s', 2) 
                ''' % {'table':table, 'name':name, 'srid':srid, 'type':type})  
        #self._create_spatial_index(table,name)
                                
    def create_table(self, table, col_specs):
        col_specs = colspecs.fix(col_specs)
        self.exec_sql('DROP TABLE IF EXISTS %s' % table)

        data_specs = []  
        geom_specs = []
        for col,typ,prop in col_specs:
            if typ=='LINESTRING' or typ=='POINT':
                geom_specs.append((col,typ,prop))
            else:
                data_specs.append((col,typ,prop))
        spec_str = [ '"%s" %s' % spec[:2] for spec in data_specs ]
        spec_str = ', '.join(spec_str)

        self.exec_sql('CREATE TABLE %s (%s)' % (table, spec_str))

        # add geometry column(s)
        for col_name, col_type, parm_name in geom_specs:
            self.add_geometry_column(table,col_name,col_type)

    def write_row(self, table, col_specs, row, srid=None):
        if not srid: srid=self.srid
        col_specs = colspecs.fix(col_specs)

        def code_value(col_name, col_type, value, ref):
            if col_type == 'POINT':
                return (self.code_point(value,srid),None)
            if col_type == 'LINESTRING':
                return (self.code_linestring(value,srid),None)
            if col_type == 'TEXT':
                if not isinstance(value,basestring): value = str(value)
                return (ref, value)
            return (ref, value)

        # create row dictionary
        if isinstance(row,dict):
            rowd = row.copy()
        else:
            # is this even used?
            assert len(row) == len(col_specs)
            rowd = {}
            for i in range(len(col_specs)): rowd[col_specs[i][2]] = row[i]

        # col and value list strings
        columns = []
        values = []
        parms = []
        for col_name,col_type,prop_name in col_specs:
            if not prop_name in rowd.keys(): continue
            value = rowd[prop_name]
            if value == None: continue
            columns.append( '"' + col_name + '"')
            (value,parm) = code_value(col_name, col_type, value, self._interpolation_ref)
            values.append(value)
            if parm!=None: parms.append(parm)
            
        columns = ','.join(columns)
        values = ','.join(values)

        # create sql template for row
        sql_template  = ''' 
                    INSERT INTO %(table)s 
		        (%(columns)s)
                    VALUES
		        (%(values)s);
                    ''' % {
                        'table':table,
                        'columns':columns,
                        'values': values
                        }

        # and execute sql
        #print 'sql_template:', sql_template
        self.exec_sql(sql_template,parms)

    def write_table(self, table, col_specs, rows):
        self.create_table(table, col_specs)

        for row in rows:
            self.write_row(table, col_specs, row)
            
    def write_geo(self, geo, table_name, srid=None, geometry_type=None, col_specs=None):
        try: geo = geo.__geo_interface__
        except AttributeError: pass
        if not geometry_type: geometry_type = geo['features'][0]['geometry']['type']
        if not col_specs: col_specs = colspecs.gen(geo, geometry_type)

        #filter features for correct geometry_type
        rows = []
        for feature in geo['features']:
            if feature['geometry']['type'] != geometry_type: continue
            row = {}
            row.update(feature['properties'])
            row['geometry'] = feature['geometry']['coordinates']
            rows.append(row)
        #print 'DEBUG rows', rows

        # write the table
        self.write_table(table_name, col_specs, rows)
        self.commit()

    def write_pnwk(self, pnwk, name=None):
        if not name: name = pnwk.name
        self.write_geo(pnwk, name+'_pnwk_segs', geometry_type='LineString')
        self.write_geo(pnwk, name+'_pnwk_jcts', geometry_type='Point')

    def add_spatial_ref_sys(self, srs_info):
        existing = self.select(
            '*',
            'spatial_ref_sys',
            where='srid=%(srid)s'% srs_info
            )
        assert len(existing) <2
        if len(existing) == 1:
            e = existing[0]
            match = True
            for col in e.keys():
                if e[col] != srs_info[col]:
                    match = False
            if not match:
                msg = 'spatial ref entry for srid=%d already exists and differs!' % srs_info['srid']
                print 'dbinterface.py WARNING:', msg
        else:
            assert len(existing) == 0
            print 'dbinterface.py:', 'Adding spatial reference system (srid=%d) to postgis database %s' % (
                srs_info['srid'],
                db_name)
            colspecs = [ (col,'generic') for col in ('srid','auth_name','auth_srid','proj4text','srtext')]
            write_row('spatial_ref_sys', colspecs, srs_info)
            commit()

    def test(self,verbose=True):
        ref = (self._interpolation_ref,)
	self.exec_sql('drop table if exists test_db')
        self.exec_sql('create table test_db ("a:f1" text, other integer, test_key serial primary key)')
        self.exec_sql('insert into test_db ("a:f1", other) values (%s, %s)' % (ref*2), 
                ('''foo's"''', 1))
        self.exec_sql('''insert into test_db ("a:f1", other) values ('bar', 5)''')
        self.exec_sql('''insert into test_db ("a:f1") values ('zar')''')
	table = self.get_table('test_db')

	assert len(table)==3
	assert table[1]['other']==5
        assert table[2]['a:f1']=='zar'
	assert len(table) == self.get_num_rows('test_db')
        col_names = self.get_column_names('test_db')
        assert col_names == ['a:f1', 'other', 'test_key']
        # another way to get col_names (but in random order)
        assert len(table[0].keys()) == 3

        #add_geometry_column
        self.add_geometry_column('test_db','geometry','POINT')
        assert self.get_column_names('test_db') == ['a:f1', 'other', 'test_key', 'geometry']

        # write_table        
        col_specs = [('id', 'BIGINT'), ('a:name', 'TEXT'), ('length', 'FLOAT'), ('geometry', 'POINT')]
        rows = []
        rows.append({ 'id':1, 'a:name':'dianna', 'length':5.3})
        rows.append({ 'id':2, 'a:name':'nathan'})
        rows.append([3,'michael',5.10, (-122.0, 38.0)])
        rows.append({ 'id':4, 'a:name':None, 'length':100 })

        self.write_table('test_db2', col_specs, rows)

        assert  self.get_column_names('test_db2') == ['id', 'a:name', 'length', 'geometry']

        rowsr = self.get_table('test_db2')
        assert len(rowsr) == 4
        assert rowsr[0]['a:name'] == 'dianna'
        assert rowsr[2]['id'] == 3
        assert rowsr[1]['length'] == None

        # geo
        geometry = geojson.Point((-122.0,38.0))
        ls = geojson.LineString([(-122,38),(-122,39)])
        features = [
            geojson.Feature(geometry=geometry, id=1, properties={'name':'john','addr':1212, 'gpa':3.5}),
            geojson.Feature(geometry=geometry, id=2, properties={'fish':'salmon', 'addr':'sea', 'gpa':3, 'foo':None}),
            geojson.Feature(geometry=ls, id=10, properties={'other':10}),
            geojson.Feature(geometry=geometry, id=3, properties={'num':2**40}),
            geojson.Feature(geometry=geometry, id=4, properties={'Num':2}),
            geojson.Feature(geometry=geometry, id=5, properties={'NUM':'many'})
            ]
        geo = geojson.FeatureCollection(features)
        
        #write_geo
        self.write_geo(geo, 'test_segs', geometry_type='LineString')
        self.write_geo(geo, 'test_jcts', geometry_type='Point')
        
        # table_names
        table_names = self.get_table_names()

        #print 'table_names:',table_names
        assert 'test_db' in table_names

        # check for spatial extension
        assert 'spatial_ref_sys' in self.get_table_names()

        # check for srid
        rows = self.select('*','spatial_ref_sys','srid=%d'%self.srid)
        assert len(rows) == 1
      
        self.commit()

def test():
    print 'dbinterface PASS'

#doit
test()



'''
Path Network.
Built up via class extension:
    PNwkNamespace  - network name, attribute name prefixes (BASE CLASS)
    PNwkNetwork    - path network implementation, read, write
    PNwkScore      - match scoring
    PNwkMatchJcts
    PNwkMatchSpokes
    PNwkMatch      - matcher
    PNwk           - top level.
'''

import pnwk_match

class PNwk(pnwk_match.PNwkMatch):
    pass

def test():
    foo = PNwk('foo')
    assert foo.name == 'foo'

    print 'pnwk PASS'

#doit
test()



'''
interface to spatialite
'''

import os
import pdb
import tempfile

import pyspatialite
import pyspatialite.dbapi2 
import dbinterface

class Slite(dbinterface.DatabaseInterface):
    def __init__(self, dbName, verbose=True):
        super(Slite,self).__init__(dbName, verbose=verbose)        

    # establish database connection 
    def _connect(self):
	con = pyspatialite.dbapi2.connect(self.db_name)
        con.row_factory = dict_factory
	cur = con.cursor()
	return [con, cur]

    def _add_spatial_extension(self):
        if not 'spatial_ref_sys' in self.get_table_names():
            if self.verbose:
                print 'SELECT spatialite.py:','Initializing spatialite metadata for %s' % self.db_name
            self.exec_sql('select InitSpatialMetadata()')
            self.commit()
        assert 'spatial_ref_sys' in self.get_table_names()

    _interpolation_ref = '?'

    def get_column_names(self, table):
        rows = self.exec_sql_r("PRAGMA table_info(%s)" % table)
        return[row['name'] for row in rows]

    def get_table_names(self):
        rows = self.exec_sql_r("select name from main.sqlite_master WHERE type='table'")
        return [ row['name'] for row in rows ]

def dict_factory(cursor, row):
    d = {}
    for idx,col in enumerate(cursor.description):
        d[col[0]] = row[idx]
    return d

def test():
    with tempfile.NamedTemporaryFile(suffix='.sqlite', delete=False) as temp:
        fname = temp.name
    try:
        #print 'fname',fname
        db = Slite(fname, verbose=False)
        db.test(verbose=False)
    finally:
        if os.path.exists(fname): os.remove(fname)
    print "spatialite PASS"

if __name__ == "__main__":
    # test is slow for some reason, so only run when invoked as main.
    test()



#TODO clean this up with ColSpecs class

import geojson

import geo

# auto gen colspecs from geo features
def gen(geo, geometry_type=None):
    geo_type_table = {'LineString':'LINESTRING', 'Point':'POINT'}
    type_name = { int:'INT', float:'FLOAT', basestring:'TEXT'}
    type_rank = { int:1, float:2, basestring:3 }

    def make_bigint_specs(specs, features, geometry_type):
        for (prop,type_name) in specs.items():
            if type_name != 'INT': continue
            for feature in features:
                if feature['geometry']['type'] != geometry_type: continue
                if abs(feature['properties'].get(prop,0)) > 2**30: 
                    specs[prop] = 'BIGINT'
                    break

    if not geometry_type: geometry_type = geo['features'][0]['geometry']['type']
    specs = {}
    specs_rank = {}
    features = geo['features']

    # auto create 'col types' based on python types of values
    # if attribute values not of same type, promote based on type_rank
    for feature in features:
        if feature['geometry']['type'] != geometry_type: continue
        for (prop,value) in feature['properties'].items():
            if value==None: continue
            t = type(value)
            if isinstance(value,basestring): t = basestring # includes unicode strings.
            if t not in type_rank: t = basestring # anything weird goes to TEXT.
            if specs_rank.get(prop,0) >= type_rank[t]: continue        
            specs[prop] = type_name[t]
            specs_rank[prop] = type_rank[t]

    # promote 'INT' attributes with large values to 'BIGINT'
    make_bigint_specs(specs, features, geometry_type)
    col_specs = sorted(specs.items())
    
    # add geometry as last column
    col_specs.append(('geometry', geo_type_table[geometry_type]))
    
    return col_specs

# uniquifies col names to avoid collisions due to case-insenstivity, and fills out prop_names
def fix(col_specs):
    def suffix (i): 
        if i==0: return ''
        return '$%d' % i
    lc_names = {}
    new_col_specs = []
    for col_spec in col_specs:
        col_name = col_spec[0]
        if len(col_spec) >= 2:
            col_type = col_spec[1]
        else:
            col_type = 'TEXT'
        if len(col_spec) >=3:
            prop_name = cols_spec[2]
        else:
            prop_name = col_name

        i=0
        while col_name.lower() + suffix(i) in lc_names: i += 1
        lc_names[col_name.lower() + suffix(i)] = True 
        new_col_specs.append((col_name + suffix(i), col_type, prop_name))
        
    return new_col_specs

def _make_geo():
    geometry = geojson.Point((-122.0,38.0))
    ls = geojson.LineString([(-122,38),(-122,39)])
    features = [
        geojson.Feature(geometry=geometry, id=1, properties={'name':'john','addr':1212, 'gpa':3.5}),
        geojson.Feature(geometry=geometry, id=2, properties={'fish':'salmon', 'addr':'sea', 'gpa':3, 'foo':None}),
        geojson.Feature(geometry=ls, id=10, properties={'other':10}),
        geojson.Feature(geometry=geometry, id=3, properties={'num':2**40}),
        geojson.Feature(geometry=geometry, id=4, properties={'Num':2}),
        geojson.Feature(geometry=geometry, id=5, properties={'NUM':'many'})
        ]
    geo = geojson.FeatureCollection(features)
    return geo

def test():
    geo = _make_geo()
    col_specs = gen(geo)
    col_specs = fix(col_specs)
    #print 'col_specs',col_specs
    assert col_specs == [('NUM', 'TEXT', 'NUM'), ('Num$1', 'INT', 'Num'), 
            ('addr', 'TEXT', 'addr'), ('fish', 'TEXT', 'fish'), ('gpa', 'FLOAT', 'gpa'), 
            ('name', 'TEXT', 'name'), ('num$2', 'BIGINT', 'num'), ('geometry', 'POINT', 'geometry')]

    col_specs = gen(geo,geometry_type='LineString')
    #print 'col_specs',col_specs
    assert col_specs == [('other','INT'), ('geometry','LINESTRING')]

    print "colspecs PASS"

#doit
test()




'''
interface to postgis
'''

import pdb

import psycopg2
from psycopg2.extras import DictCursor

import dbinterface

class PGIS(dbinterface.DatabaseInterface):
    def __init__(self, dbName, verbose=True):
        super(PGIS,self).__init__(dbName, verbose=verbose)        

    # establish database connection 
    # NOTE: postgres server must be running and database must already exist 
    # Initial database creation (from shell prompt):
    #   % createdb <db_name>
    def _connect(self):
	con = psycopg2.connect(database=self.db_name)
	cur = con.cursor(cursor_factory=psycopg2.extras.DictCursor)
	return [con, cur]

    def _add_spatial_extension(self):
        if not 'spatial_ref_sys' in self.get_table_names():
            if self.verbose: 
                print 'pgis.py:','Adding postgis extension to postgres database %s' % self.db_name
            self.exec_sql('create extension postgis;')
            self.commit()
        assert 'spatial_ref_sys' in self.get_table_names()

    _interpolation_ref = '%s'

    def get_column_names(self,table):
       name_per_row = self.select('column_name', 
                 'information_schema.columns',
                  where="table_name='%s'"%table)
       col_names=[]
       for row in name_per_row: col_names.append(row[0])
       return col_names

    def get_table_names(self):
        rows = self.select('table_name','information_schema.tables')
        return [ row[0] for row in rows ]

def test():
    # requires preexisting spaitially enabled 'test' database:
    # %createdb test
    pgis = PGIS('test', verbose=False)
   
    # run tests
    pgis.test(verbose=False)

    print "postgis PASS"

# on module load
test()




'''
Street Centerline Data to Path Network conversion
'''
import pdb

import geo
import pnwk

''' features in python geo interface format (geojson) '''
def make_pnwk(features, props=None, namesFunc=None, filter_func=None, name=None,quiet=False, clip_rect=None):
    if not quiet: print 'num centerline features:', len(features)
    #print 'DEBUG random city feature:',features[10]

    city_street_network = pnwk.PNwk(name=name)
    num_segs = 0
    num_jcts=[0]  # this is a hack to get around scoping issues
    jct_ids = {}  # indexed by jct coords
    skipped_features = []
    max_input_dim = 0

    # return jct_id for point (assigning new one if needed)
    def jct_id(point):
        if jct_ids.has_key(point):
            return jct_ids[point]
        else:
            num_jcts[0] += 1 
            jct_ids[point] = num_jcts[0]
            return num_jcts[0]

    for feature in features:
        if filter_func and not filter_func(feature): continue

        geometry = feature['geometry']
        if geometry['type'] == 'MultiLineString':
            parts = geometry['coordinates']
        else:
            if geometry['type'] == 'LineString':
                parts = [geometry['coordinates']]
            else:
                parts = []
                skippedFeatures.append(feature)

        for part in parts:
            num_segs += 1
            seg_id = num_segs
            points = []
            for l in part:
                #code assumes 2d coords, retaining only first 2 coords of input.
                points.append(tuple(l[:2]))
                max_input_dim = max(max_input_dim,len(l))
            start_point = points[0];
            end_point = points[-1]

            if not props:
                tags = feature['properties']
            else:
                tags = {}
                for prop in props:
                    if prop in feature['properties']: 
                        tags[prop] = feature['properties'][prop]

            if clip_rect and not geo.points_intersect_rect_q(points, clip_rect): continue 

            city_street_network.add_seg(seg_id, jct_id(start_point), jct_id(end_point), points,
                    names=namesFunc(feature),
                    tags=tags
                    )
   
    if len(skipped_features) > 0  and not quiet:
        print '%d features skipped.' % len(skipped_features)
        print 'first skipped feature:',feature

    if max_input_dim > 2 and not quiet:
        print 'WARNING: Input coordinates up to %d dimensional.' % max_input_dim
        print 'Retaining only first 2 dimensions.'

    return city_street_network

def rlis_pnwk(features,name=None,props=None,quiet=False,clip_rect=None):
    def rlis_names(feature):
        def _rlis_names0(prefix,street_name,ftype):
            #print 'DEBUG _rlis_names0 prefix,street_name,ftype', prefix,street_name,ftype

            if not street_name or street_name=='UNNAMED': return []
            names = []

            # <prefix> <street_name> <ftype>
            name = ''
            if prefix: name += (prefix + ' ')
            name += street_name
            if ftype: name += (' ' + ftype)
            names.append(name)

            # split out names in RAMPs
            if ftype == 'RAMP':
                for sname in street_name.split('-'):
                    names += _rlis_names0(prefix,sname,'')
                    names += _rlis_names0('',sname,'')  # want 'I205' not 'NE I205'
            
            #print 'DEBUG _rlis_names0 names', names 
            return names
        properties = feature['properties']
        prefix = properties.get('PREFIX')
        street_name = properties.get('STREETNAME')
        ftype = properties.get('FTYPE')
        return _rlis_names0(prefix,street_name,ftype)
    return make_pnwk(features,
            props=props,
            namesFunc=rlis_names,
            name=name,
            quiet=quiet,
            clip_rect=clip_rect)

def berkeley_pnwk(features, name=None, props=None, quiet=False, clip_rect=None):
    def berkeley_names(feature): 
        names = []
        props = feature['properties']
        cat = props['CATEGORY']
        name = props['FULLNAME']
        words = name.split(' ')
        first = words[0]
        if name=='TRAIL' and cat =='PEDESTRIAN': return names 
        if name=='RAMP' and cat == 'CONNECTOR': return names
        if name=='ALLEY': return names
        if 'UNNAMED' in name: return names
        names.append(name)

        # add unadorned freeway names.
        if cat in ('CONNECTOR','HIGHWAY') and len(words)>1 and len(first)>1 and first[0]=='I':
            try: 
                if int(first[1:]) > 0: names.append(first)
            except ValueError: pass

        # for 'HWY 13 N', add 'CA 13'
        if cat in ('CONNECTOR','HIGHWAY') and first=='HWY' and len(words)>1:
            try: 
                if int(words[1]) > 0: names.append('CA ' + words[1])
            except ValueError: pass
        return names

    def berkeley_filter(feature):
        category = feature['properties']['CATEGORY']
        return category in ('CONNECTOR','HIGHWAY','MAJOR','MINOR','PEDESTRIAN')

    return make_pnwk(features,
            props=props,
            namesFunc=berkeley_names,
            filter_func=berkeley_filter,
            name=name,
            quiet=quiet,
            clip_rect=clip_rect)

def test():
    features = [
    {
      "geometry": {
        "type": "LineString", 
        "coordinates": [ [10,11],[20,21] ]
        }, 
      "type": "Feature", 
      "properties": {
        "OBJECTID": 1, 
        "FULLNAME": "I80 E UNIVERSITY RAMP", 
        "CATEGORY": "CONNECTOR"
      }
    },

    {
      "geometry": {
        "type": "LineString", 
        "coordinates": [ [100,110, 100],[200,210, 110] ]
        }, 
      "type": "Feature", 
      "properties": {
        "OBJECTID": 2, 
        "FULLNAME": "TRAIL", 
        "CATEGORY": "PEDESTRIAN"
      }
    },

    {
      "geometry": {
        "type": "LineString", 
        "coordinates": [ [100,110, 100],[200,210] ]
        }, 
      "type": "Feature", 
      "properties": {
        "OBJECTID": 10, 
        "FULLNAME": "filter me",
        "CATEGORY": "LIGHT RAIL"
      }
    }]

    pnwk = berkeley_pnwk(features,quiet=True)
    assert len(pnwk.segs) == 2
    names = pnwk.segs[1].names
    assert len(names) == 2
    assert 'I 80 E UNIVERSITY RAMP' in names
    assert 'I 80' in names
    assert len(pnwk.segs[2].names_text()) == 0

    # test clipping
    clip_rect = [ 15,15,25,25 ] 
    pnwk = berkeley_pnwk(features,quiet=True,clip_rect=clip_rect)
    assert len(pnwk.segs) == 1
    assert 'I 80 E UNIVERSITY RAMP' in pnwk.segs[1].names

    print 'centerline PASS'

#doit 
test()



'''
Geography / Geometry utilities.
'''

import math
import cmath
#import numpy

import pyproj
import shapely
import shapely.geometry

class Projection(object):
    def __init__(self,proj4text):
        # longlat = unprojected  ESPG:4326  (corresponds to OSM data)
        self.ll_proj = pyproj.Proj('+proj=longlat +ellps=WGS84 +datum=WGS84 +no_defs')
        self.proj = pyproj.Proj(proj4text, preserve_units=True)

    # latlon -> local coords
    def project_point(self,p, rev=False):
        if rev: return pyproj.transform(self.proj, self.ll_proj, p[0], p[1])
        return pyproj.transform(self.ll_proj, self.proj, p[0], p[1])
  
    def project_box(self,box, rev=False):
        xmin,ymin,xmax,ymax = box
        pmin = self.project_point((xmin,ymin), rev=rev)
        pmax = self.project_point((xmax,ymax), rev=rev)
        return pmin + pmax
	
    # geo projections done in place
    def project_geo_linestring(self,ls, rev=False):
        new = []
        for p in ls['geometry']['coordinates']:
            new.append(self.project_point(p, rev=rev))
        ls['geometry']['coordinates'] = new

    def project_geo_features(self, geo, rev=False):
        for feature in geo:
            t = feature['geometry']['type']

            if t=='LineString': 
                self.project_geo_linestring(feature, rev=rev)
            else:
                # TODO other types. :)
                assert False

def _test_projection():
    proj = Projection('+proj=utm +zone=10 +ellps=WGS84 +units=m +no_defs')
    sf_geo = (-122.0,38.0)
    sf_utm = proj.project_point(sf_geo)
    sf_geo_roundtrip = proj.project_point(sf_utm, rev=True)
    epsilon = .000001
    #print 'sf_geo,sf_utm,sf_geo_roundtrip', sf_geo, sf_utm, sf_geo_roundtrip
    assert abs(sf_geo[0]-sf_geo_roundtrip[0]) < epsilon
    assert abs(sf_geo[1]-sf_geo_roundtrip[1]) < epsilon

# WKT
# Could use Shapely for this?
def wkt_point(point):
	return 'POINT (%.6f %.6f)' % point
def wkt_linestring(points):
	assert len(points)>1
	a = []
	for p in points: a.append('%.6f %.6f' % p)
	return 'LINESTRING (' + ', '.join(a) + ')'
def wkt_linestring_to_points(wkt_string):
    rx = re.compile(r"^\w*LINESTRING\w*\((.*)\)\w*$", re.IGNORECASE)
    point_string = rx.search(wkt_string).group(1)       
    coords = re.split(r"\s*,?\s*",point_string)
    assert len(coords)%2 == 0  # number of coords should be even!
    points=[]
    for i in xrange(0,len(coords),2):
        point = ( float(coords[i]), float(coords[i+1]) )
        points.append(point)
    return points

def shapely_point(point):
    return shapely.geometry.Point(point)

def shapely_linestring(points):
    return shapely.geometry.LineString(points)

def shapely_point_to_point(shapely_point):
    point_list = list(shapely_point.coords)
    assert len(point_list) == 1
    return point_list[0]
    
def shapely_linestring_to_points(shapely_linestring):
    return list(shapely_linestring.coords) 

def _test_shapely():
    p = (5,6.5)
    sp = shapely_point(p)
    pout = shapely_point_to_point(sp)
    assert p == pout

    ls = [(0,0),(1,10),(2,20)]
    sls = shapely_linestring(ls)
    lsout = shapely_linestring_to_points(sls)
    assert ls == lsout


# GEOMETRY

def buffer_box(box, amount):
    xmin,ymin,xmax,ymax = box
    xmin -= amount; ymin -= amount
    xmax += amount; ymax += amount
    return (xmin,ymin,xmax,ymax)

def distance(p0,p1):
	deltax= p1[0] - p0[0]
	deltay= p1[1] - p0[1]
	return abs(complex(deltax,deltay))

def _test_distance():
	assert distance((5,0),(15,0)) == 10
	assert distance((1,2),(4,6)) == 5

def length(points):
    d = 0
    prev = None
    for point in points:
        if prev: d += distance(prev,point)
        prev = point
    return d

def _test_length():
    assert length([]) == 0
    assert length([(5,10)]) == 0
    assert length([(0,0),(0,1)]) == 1
    assert length([(0,0),(0,1),(1,1),(1,0)]) == 3

# point along linestring distance d from beginning.
def interpolate(points,d):
    sls = shapely_linestring(points)
    spoint = sls.interpolate(d)
    return shapely_point_to_point(spoint)

def _test_interpolate():
    ls = [(0,0),(0,10),(10,10),(10,0),(0,0)]
    assert interpolate(ls, 5) == (0.0,5.0)
    assert interpolate(ls, 20) == (10.0,10.0)
    assert interpolate(ls, 39) == (1.0,0.0)

# returns (maxSep,avg_bearing_delta)
def divergence(points1, points2, delta=3):
    len1 = length(points1)
    len2 = length(points2)

    if len1<len2: 
        return divergence(points2,points1, delta=delta)
    assert len1 >= len2

    sls1 = shapely_linestring(points1)
    sls2 = shapely_linestring(points2)
    try:
        f = len2/len1
    except ZeroDivisionError:
        assert len1==0 and len2==0
        f= 1
    d1=0
    maxSep = 0
    p1_prev = None
    p2_prev = None
    num_b = 0
    cum_bdelta = 0
    while True:
        sp1 = sls1.interpolate(d1)
        sp2 = sls2.interpolate(d1*f)
        p1 = shapely_point_to_point(sp1)
        p2 = shapely_point_to_point(sp2)
        sep = sp1.distance(sp2)
        maxSep = max(maxSep,sep)

        if p1_prev:
            b1 = compass_bearing(p1_prev,p1)
            b2 = compass_bearing(p2_prev,p2)
            bdelta = compass_bearing_delta(b1,b2)
            cum_bdelta += bdelta
            num_b += 1 
        p1_prev = p1
        p2_prev = p2

        if d1==len1: break
        d1 += delta
        d1 = min(d1,len1)

    #print 'cum_bdelta', cum_bdelta
    #print 'num_b', num_b
    try:
        avg_bdelta = (cum_bdelta+0.0)/num_b
    except ZeroDivisionError:
        avg_bdelta = 0
    return (maxSep, avg_bdelta)

def _test_divergence():
    points1 = [(0,10),(100,15),(200,10)]
    points2 = [(0,-10),(100,-15),(200,-10)]
    points3 = [(0,0),(200,0)]
    points4 = [(0,0),(400,0)]
    assert 29 < divergence(points1,points2)[0] <= 30 
    assert divergence(points1,points1)[0] == 0
    assert 14 < divergence(points3,points1)[0] <= 15
    assert divergence(points3,points4)[0] == 200

    points5 = [(0,0),(20,20),(40,0),(400,0)]
    assert divergence(points3,points4)[1] == 0
    assert 5.7 < divergence(points4,points5,delta=10)[1] < 5.8

    points6 = [(0,0), (0,0)]
    points7 = [(10,0), (10,0)]
    assert divergence(points6,points7)[0] == 10 
    assert divergence(points6,points7)[1] == 0 

# distance along points nearest to point
def project(points,point):
    sls = shapely_linestring(points)
    sp = shapely_point(point)
    return sls.project(sp)

def _test_project():
    ls = [(0,0),(0,10),(10,10),(10,0),(0,0)]
    assert project(ls,(1,5)) == 5.0
    assert project(ls,(11,11)) == 20.0

def cut(points,d):
    assert 0 < d < length(points) 
    assert len(points)>=2
    part1 = []
    part2 = []
    curd = 0
    prev= None
    for p in points:
        if prev: curd += distance(prev,p)
        if curd<=d:
            part1.append(p)
        else:
            part2.append(p)
        prev = p

    if length(part1) != d:
        newp = interpolate(points,d)
        part1.append(newp)
        part2.insert(0,newp)
    else:
        part2.insert(0,part1[-1])

    return part1,part2

def _test_cut():
    points = [(0,1),(10,1),(20,1),(30,1)]
    part1,part2 = cut(points,5.0)
    #print 'part1:', part1
    #print 'part2:', part2
    assert len(part1)==2 and len(part2)==4
    part1,part2 = cut(points, 20.0)
    #print 'part1:', part1
    #print 'part2:', part2
    assert len(part1)==3 and len(part2)==2

def point_in_rect_q(point, rect):
    px,py = point
    x0,y0,x1,y1 = rect
    return x0<=px<x1 and y0<=py<y1

def _test_point_in_rect_q():
    assert point_in_rect_q((1,2),(0,0,10,10))
    assert not point_in_rect_q((100,2),(0,0,10,10))

# does at least one of the points fall inside rect?
def points_intersect_rect_q(points, rect):
    for point in points:
        if point_in_rect_q(point, rect): return True
    return False

def _test_points_intersect_rect_q():
    assert not points_intersect_rect_q([(100,1),(200,2),(300,3)], (0,0,10,10)) 
    assert points_intersect_rect_q([(100,1),(2,2),(300,3)], (0,0,10,10)) 

def compass_bearing(p0,p1):
	deltax = p1[0] - p0[0]
	deltay = p1[1] - p0[1]
	theta = math.degrees(cmath.phase(complex(deltax,deltay)))
	bearing = 90-theta
	if bearing<0 : bearing += 360
	return bearing

def compass_bearing_delta(bearing0, bearing1):
    delta = bearing0-bearing1
    delta = abs(delta) % 360
    if delta>180: delta = 360-delta
    #print 'cbd:', bearing0, bearing1, delta
    return delta

def _test_compass_bearing():
	cb=compass_bearing
	assert cb((0,0),(0,1)) == 0
	assert cb((0,0),(10,0)) == 90
	assert cb((0,0),(0,-110)) == 180
	assert cb((0,1),(-11,1)) == 270
	assert cb((5,10),(5,20)) == 0

def _test_compass_bearing_delta():
    bd= compass_bearing_delta
    assert bd(350,10) == 20
    assert bd(10,350) == 20
    assert bd(-90,90) == 180
    assert bd(0,340) == 20

class BBox(object):
    def __init__(self):
        self.box = None

    def rect(self):
        return self.box

    def add_point(self,point):
        x,y = point
        if not self.box:
            self.box = (x,y,x,y)
        else:
            x0,y0,x1,y1 = self.box
            self.box = (min(x0,x),min(y0,y),max(x1,x),max(y1,y))

def _test_BBox():
    bb = BBox()
    assert bb.rect() == None
    bb.add_point((10,100))
    assert bb.rect() == (10,100,10,100)
    bb.add_point((-5,120))
    assert bb.rect() == (-5, 100, 10, 120)

def _test_linestring_to_points():
    ls = 'LINESTRING(1.1,10.0 2.2, 20.0)'
    points = wkt_linestring_to_points(ls)
    assert len(points) == 2
    assert points[1][0] == 2.2 and points[1][1] == 20.0

def _test_linestring_conversions():
    wkt_in = "LINESTRING(7696992.83333333 682678.163713917,7696997.45603675 682808.892060369)"
    print 'wkt_in:', wkt_in
    points = wkt_linestring_to_points(wkt_in)
    print 'points:', points
    wkt_out = wkt_linestring(points)
    print 'wkt_out:', wkt_out

def test():
        #test_linestring_to_points()
        _test_projection()
	_test_distance()
        _test_length()
        _test_divergence()
	_test_compass_bearing()
        _test_compass_bearing_delta()
        _test_point_in_rect_q()
        _test_points_intersect_rect_q()
        _test_BBox()
        _test_shapely()
        _test_interpolate()
        #_test_project()
        _test_cut()
	print 'geo PASS'

#doit
test()




'''
Path Network - Match 'spokes,' crawling outward from already matched jcts, and splitting longer segments as needed.
'''
import pdb

import geo 
import pnwk_matchjcts

class PNwkMatchSpokes(pnwk_matchjcts.PNwkMatchJcts):
    class Jct(pnwk_matchjcts.PNwkMatchJcts.Jct):
        def spokes(self):
            spokes=[]
            for seg in self.segs:
                spokes.append(Spoke(jct=self, seg=seg))
            return spokes

    # match nwk2 into nwk1
    # matches are searched for in radius d and must be unique within that radius in both networks
    def spoke_crawl(self, star_jct_match_list, msg=None, quiet=False):
        #if msg and not quiet: print 'SPOKECRAWL:',msg
        new_jct_matches = star_jct_match_list
        passNum = 0
        while len(new_jct_matches)>0:
            passNum += 1
            #if not quiet: 
            if False:
                print 'pass %d, processing %d matched junctions:' % (passNum, len(new_jct_matches))
            new_jct_matches = match_spokes(new_jct_matches)

        return passNum>1

class Spoke(object):
    def __init__(self, jct, seg):
        assert jct.network == seg.network
        self.jct = jct
        self.seg = seg

        if seg.from_jct == jct:
            self.bearing = seg.from_bearing
        else:
            assert seg.to_jct == jct
            self.bearing = seg.to_bearing

    # is spoke direction reversed from segment direction
    def upstream(self):
        return self.seg.to_jct == self.jct

    # return jct at far end of spoke
    def end(self):
        return self.seg.other_end(self.jct)

    # return spoke that continues on from far end of this spoke (if any)
    def next(self):
        jct = self.end()
        bestDelta = 180
        bestSpoke = None;
        for spoke in jct.spokes():
            delta = geo.compass_bearing_delta(self.bearing,spoke.bearing)
            if delta<bestDelta: 
                bestSpoke = spoke
                bestDelta = delta
        if bestDelta > 90: return None
        return bestSpoke

    # return point at distance d along spoke
    def interpolate(self,d):
        if self.upstream(): 
            deff = self.length() - d
        else:
            deff = d

        assert -0.000001 <= d and d <= self.length() + .000001
        return geo.interpolate(self.seg.points,deff)

     # return distance along spoke to nearest approach to point
    def project(self,point):
        d = geo.project(self.seg.points,point)
        if self.upstream(): d = self.length() - d
        return d

    def length(self):
        return self.seg.length()

    # trim spoke length (by splitting seg)
    def trim(self,d,new_jct_id=None):
        assert 0<d and d<self.length()
        fromBeginning = not self.upstream()
        self.seg.split(d, fromBeginning, new_jct_id=new_jct_id) 
        assert not new_jct_id or self.end().jct_id==new_jct_id

    # return subset of spokes that match self (along with distances along spokes of match)
    def match(self, spokes):
        match_set = []
        if self.seg.match: return match_set

        len1 = self.length()
        street_scale = self.seg.network.street_scale
        for spoke in spokes:
            if spoke.seg.match: continue
            len2 = spoke.length()

            if len1<=len2:
                length = len1
                d1=length
                d2=spoke.project(self.end().point)
            else:
                length = len2
                d1 = self.project(spoke.end().point)
                d2 = length

            # to match, segments must be approximately same length
            if abs(d1-d2) > max(street_scale, 0.1*length): continue

            # to match, segment end points must be near each other
            if geo.distance(self.interpolate(d1),spoke.interpolate(d2)) > street_scale: continue

            match_set.append((self, d1, spoke, d2))

        return match_set
 
def _trim_spoke_q(shorter, longer, d):

    # if length difference significant, trim
    if abs(longer.length()-d)>longer.seg.network.street_scale:
        return True 
   
    # if next point after shorter spoke end is better match for longer end, trim!
    next_shorter = shorter.next()
    if next_shorter:
        delta_next = geo.distance(longer.end().point, next_shorter.end().point)
        delta_this = geo.distance(longer.end().point, shorter.end().point)
        return delta_next < delta_this

    # no good reason to trim
    return False
      
# consider trimming longer spoke.
def trimMatchedSpokes(spoke1, d1, spoke2, d2):
    length1 = spoke1.length()
    length2 = spoke2.length()

    # make spoke1 the short one.
    if(length1>length2):
        trimMatchedSpokes(spoke2, d2, spoke1, d1)
        return
    assert length1<=length2
    assert length1==d1

    shorter = spoke1
    longer = spoke2
    d = d2

    # stay away from ends
    middle = length2/2.0
    if d<1: d= min(1,middle)
    if length2-d<1: d = max(middle,length2-1)

    long = spoke2
    if _trim_spoke_q(shorter, longer, d):
        longer.trim(d)
        shorter.end().tags['export_count'] = shorter.end().tags.get('export_count',0) + 1
        longer.end().tags['import_src'] = shorter.end()

def _match_jct_spokes(jct1, jct2):
    global there
    assert jct1.network != jct2.network
    new_jct_matches = []
    
    spokes1 = jct1.spokes()
    spokes2 = jct2.spokes()
    #print 'len1, len2:', len(spokes1), len(spokes2)

    for spoke1 in spokes1:
        match_set = spoke1.match(spokes2)
        
        #matches must be unique
        if len(match_set) != 1: continue

        spoke1x,d1,spoke2,d2 = match_set[0]
        assert spoke1x == spoke1

        #matches must be unique in both directions
        match_set2 = spoke2.match(spokes1)
        assert spoke1 in [ m[2] for m in match_set2 ]
        if len(match_set2) != 1: continue

        assert not spoke1.seg.match
        assert not spoke2.seg.match

        # consider trimming longer spoke
        trimMatchedSpokes(spoke1, d1, spoke2, d2)

        # match spoke segs
        #jct1.network.check() #DEBUG!
        #jct2.network.check()
        assert not spoke1.seg.match
        assert not spoke2.seg.match
        spoke1.seg.match = spoke2.seg
        spoke2.seg.match = spoke1.seg
        #jct1.network.check()
        #jct2.network.check()

        # sense of matched segs same or reverse?
        if spoke1.upstream() == spoke2.upstream():
            rev = 0
        else:
            rev = 1
        spoke1.seg.match_rev = rev
        spoke2.seg.match_rev = rev

        # continue spoke crawl where we left off.
        new_jct_matches.append((spoke1.end(), spoke2.end()))
        there = False

    return new_jct_matches

def match_spokes(jct_match_list):
    new_jct_matches = []
    for (jct1,jct2) in jct_match_list:
        new_jct_matches += _match_jct_spokes(jct1, jct2)
    return new_jct_matches


def test():
    points = [(0,0), (1000,1000)]
    points2 = [(5,0), (1005,1000)]
    g = PNwkMatchSpokes('g')    
    g.add_seg(1, 10, 11, points, names=['foo',
        "NE HALSEY ST FRONTAGE RD",
        "Northeast Halsey Street Frontage Road"])

    # test spokes
    spokes = g.jcts[10].spokes()
    assert len(spokes)==1 and spokes[0].bearing==45 and spokes[0].seg.seg_id==1 
    spokes = g.jcts[11].spokes()
    assert len(spokes)==1 and spokes[0].bearing==225 and spokes[0].seg.seg_id==1 
    g.check()

    # test segment and spoke length
    n3 = PNwkMatchSpokes('n3_nwk')
    points3 = [(0,0),(1,0),(1,1)]
    n3.add_seg(1, 10, 11, points3)
    assert n3.segs[1].length() == 2
    assert n3.jcts[11].spokes()[0].length() == 2

    # test spoke.end()
    assert n3.jcts[10].spokes()[0].end() == n3.jcts[11]
    assert n3.jcts[11].spokes()[0].end() == n3.jcts[10]

    # test spoke.upstream()
    spoke_out = n3.jcts[10].spokes()[0]
    spoke_in = n3.jcts[11].spokes()[0]
    assert not spoke_out.upstream()
    assert spoke_in.upstream()

    # test spoke.interpolate()
    assert spoke_out.interpolate(0) == (0,0)
    assert spoke_out.interpolate(1.5) == (1,0.5)
    assert spoke_in.interpolate(0) == (1,1)
    assert spoke_in.interpolate(1.5) == (0.5,0)

    # test spoke.project()
    assert spoke_out.project((10,0.5)) == 1.5
    assert spoke_in.project((10,0.5)) == 0.5

    # test spoke.trim()
    spoke_in.trim(0.5,-1)
    spoke_out = n3.jcts[10].spokes()[0]
    spoke_out.trim(0.5,-2)
    assert len(n3.jcts) == 4
    assert len(n3.segs) == 3
    assert spoke_in.end().point == (1,0.5)
    assert spoke_out.end().point == (0.5,0)

    # test spoke.next()
    nstar = PNwkMatchSpokes('star_nwk')
    nstar.add_seg(360,0,360,[(0,0),(0,100)])
    nstar.add_seg(90,0,90,[(0,0),(100,0)])
    nstar.add_seg(180,0,180,[(0,0),(0,-100)])
    nstar.add_seg(270,0,270,[(0,0),(-100,0)])
    assert Spoke(seg=nstar.segs[360],jct=nstar.jcts[360]).next().end().point == (0, -100)
    assert Spoke(seg=nstar.segs[270],jct=nstar.jcts[270]).next().end().point == (100, 0)
    assert Spoke(seg=nstar.segs[270],jct=nstar.jcts[0]).next() == None

    print 'pnwk_matchspokes PASS'

#doit
test()




'''
Base class for PNwk (Path Network)
'''

class PNwkNamespace(object):

    NAMESPACE_SEP = '$'

    def __init__(self,name=None):

        #network name
        if not name: name = 'nwk'
        self.name = name

        # namespace prefixes for properties
        sep = self.NAMESPACE_SEP
        self.client_name_space = self.make_namespace(name)
        self.pnwkNameSpace = self.make_namespace(name + '_pnwk') 
        self.match_namespace = self.make_namespace('match')

    def make_namespace(self,name):
        return name + self.NAMESPACE_SEP

    def add_namespace(self, props, nameSpace):
        if not props: return {}
        out = {}
        for key in props.keys():
            out[nameSpace+key] = props[key]
        return out

    def split_off_namespace(self,name):
        i = name.find(self.NAMESPACE_SEP)+1
        return (name[:i],name[i:])

def test_add_namespace(nsp):
    a = {'id':1, 'name':2, 'length':3}
    ns = nsp.make_namespace('oz')
    b = nsp.add_namespace(a, ns)
    assert b['oz' + nsp.NAMESPACE_SEP + 'name'] == 2

def test_split_off_namespace(nsp):
    (ns,base) = nsp.split_off_namespace('foo$bar')
    assert ns == 'foo$'
    assert base == 'bar'
    (ns,base) = nsp.split_off_namespace('bar')
    assert ns == ''
    assert base == 'bar'
    (ns,base) = nsp.split_off_namespace('foo$bar$zar')
    assert ns == 'foo$'
    assert base == 'bar$zar'

def test():
    nsp = PNwkNamespace(name='fish')
    test_add_namespace(nsp)
    test_split_off_namespace(nsp)
    print 'pnwk_namespace PASS'

#doit
test()



import copy
import numbers
import pdb

import geojson

import geo

def _geo_features(geo):
    features = geo
    try: features = geo.__geo_interface__
    except AttributeError: pass
    try: features = features['features']
    except TypeError: pass
    return features

def filter_features(features, feature_func=None, geom_type=None, col_specs=None):
    features = _geo_features(features)

    new_features = []
    for feature in features:
        if geom_type and feature['geometry']['type'] != geom_type: continue
        new_feature = copy.deepcopy(feature)
        if feature_func:
            # feature_func may edit feature
            if not feature_func(new_feature,new_feature['properties']): continue
        new_features.append(new_feature)

        if col_specs:
            props = new_feature['properties']
            out = {}
            for (col_name,col_type,prop_name) in col_specs:
                if prop_name in props:
                    out[col_name] = props[prop_name]
            new_feature['properties'] = out

    return new_features

# allows for nested cooordinate lists
def coordinates_intersect_rect_q(coords, rect):
    if isinstance(coords[0][0], numbers.Number):
        return geo.points_intersect_rect_q(coords, rect)
    else:
        # nested coord lists
        for l in coords:
            if coordinates_intersect_rect_q(l, rect): return True
        return False

def _test_coordinates_intersect_rect_q():
    coords1 = [[0,0], [10,10], [20,20]]
    coords2 = [[100,100], [20,20]]
    coords_l2 = [coords1, coords2]
    assert coordinates_intersect_rect_q(coords1,[15,15,30,30])
    assert not coordinates_intersect_rect_q(coords1,[25,15,30,30])
    assert coordinates_intersect_rect_q(coords_l2,[90,90,101,101])

def test():
    _test_coordinates_intersect_rect_q()
    features = [
    {
      "geometry": {
        "type": "LineString", 
        "coordinates": [ [10,11],[20,21] ]
        }, 
      "type": "Feature", 
      "properties": {
        "OBJECTID": 1, 
        "FULLNAME": "THE UPLANDS PATH", 
        "CATEGORY": "PEDESTRIAN"
      }
    },

    {
      "geometry": {
        "type": "LineString", 
        "coordinates": [ [100,110],[200,210] ]
        }, 
      "type": "Feature", 
      "properties": {
        "OBJECTID": 2, 
        "FULLNAME": "TRAIL", 
        "CATEGORY": "PEDESTRIAN"
      }
    },

    {
      "geometry": {
        "type": "Point", 
        "coordinates": [ 100,110 ]
        }, 
      "type": "Feature", 
      "properties": {
        "OBJECTID": 10, 
        "FULLNAME": "pointy dude",
        "CATEGORY": "LIGHT RAIL STOP"
      }
    }]

    class GeoThingy(object):
        @property
        def __geo_interface__(self): return self.geo
        def __init__(self,features): 
            self.geo = geojson.FeatureCollection(features)

    # _geo_features()
    out = _geo_features(features)
    assert len(out)==3 and out[0]['type'] == 'Feature'
    out = _geo_features(geojson.FeatureCollection(features))
    assert len(out)==3 and out[0]['type'] == 'Feature'
    out = _geo_features(GeoThingy(features))
    assert len(out)==3 and out[0]['type'] == 'Feature'

    # filter_features() - trivial case
    assert len(filter_features([])) == 0
    
    # check for deep copy
    out = filter_features(features)
    assert len(out) == 3
    assert out[1]['properties']['FULLNAME'] == 'TRAIL'
    out[1]['properties']['FULLNAME'] = 'Matt Davis'
    assert features[1]['properties']['FULLNAME'] == 'TRAIL'
    assert out[1]['properties']['FULLNAME'] == 'Matt Davis'

    # geom_type
    out = filter_features(features, geom_type='Point')
    assert len(out)==1 and out[0]['geometry']['type']=='Point'

    # feature_func
    def ffunc(feature,props): 
        if not props['OBJECTID'] > 1: return False
        props['FORMAL_NAME'] = 'Ms. ' + props['FULLNAME']
        return True

    out = filter_features(features, feature_func=ffunc)
    assert len(out)==2
    assert out[0]['properties']['FORMAL_NAME'] == 'Ms. TRAIL'

    # col_specs
    specs = [('num','INT','OBJECTID'),('name','INT','FORMAL_NAME')]
    out = filter_features(features, ffunc, geom_type='LineString', col_specs=specs)
    assert len(out) == 1
    assert out[0]['properties']['name'] == 'Ms. TRAIL'

    print 'geofeatures PASS'

#doit
test()



'''
Miscellaneous shared code for spiderosm
'''

import os
import urllib
import sys
import time
import zipfile

# print immediately (no buffering)
def print_now(s, noNewLine=False):
    if noNewLine:
        print s,
    else:
        print s
    sys.stdout.flush()

def date_ymd():
    return time.strftime("%Y%m%d")         

def date_ymdhms():
    return time.strftime("%Y%m%d_%H%M%S")  

def update_file_from_url(filename,url):
    if not os.path.exists(os.path.dirname(filename)): os.makedirs(os.path.dirname(filename))

    # get current file size
    file_size = None
    if os.path.exists(filename):
        file_size = os.stat(filename).st_size

    # get size of url content
    url_size = None
    f = urllib.urlopen(url)
    i = f.info()
    url_size = int(i.getheader('content-length'))
    f.close()

    print 'file_size', file_size
    print 'url_size', url_size
    if file_size != url_size:
        print 'Downloading: %s to %s' % (url,filename)
        urllib.urlretrieve(url,filename)

def unzip(Filename):
    zfile = zipfile.ZipFile(Filename)
    (zipDir,zipFile) = os.path.split(Filename)
    for name in zfile.namelist():
        (subDir, filename) = os.path.split(name)
        dirname = os.path.join(zipDir,subDir)
        # paranoid about escaping from current dir with '..'
        assert not '..' in dirname
        assert not '//' in dirname
        assert not '\\' in dirname
        if dirname=='': dirname='.'
        print "Decompressing " + filename + " on " + dirname
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        zfile.extract(name, dirname)

def _test_dates():
    print 'date_ymd:', date_ymd()
    print 'date_ymdhms:', date_ymdhms()

def _test_print_now():
    print_now('a',noNewLine=True)
    time.sleep(2)
    print_now('b',noNewLine=True)
    time.sleep(2)
    print_now('c')
    time.sleep(5)
    print_now('d')

def test(all_=False):
    if all_: _test_dates()
    if all_: _test_print_now()
    print 'misc PASS'

#doit
test()
#test(all_=True)



'''
Testing
'''

import config
import importlib

core_modules = (
        'bins',
        'cannames',
        'centerline',
        'colspecs',
        'config',
        'csvif',
        'dbinterface',
        'geo',
        'geofeatures',
        'match',
        'misc',
        'osm',
        'pnwk',
        'pnwk_match',
        'pnwk_matchjcts',
        'pnwk_matchspokes',
        'pnwk_namespace',
        'pnwk_network',
        'pnwk_score',
        'shp2geojson',
        )

optional_modules = (
        'postgis',
        'spatialite'
        )

def test_module(name):
    if __package__:  name = __package__ + '.' + name
    print 'testing module %s...' % name
    mod = importlib.import_module(name)
    mod.test()

def test_core_modules():
    for name in core_modules:
        test_module(name)

def test_optional_modules():
    for name in optional_modules:
        if config.settings.get(name+'_enabled'): test_module(name)

def test_modules():
    test_core_modules()
    test_optional_modules()

def run_tests():
    test_modules()

#doit
if __name__ == "__main__":
    config.read_config_files()
    run_tests()
    print 'Congratulations, spiderOSM passed all test.'



import csv
import os
import tempfile

import geojson

def read_table(fileName):
    with open(fileName, 'rb') as f:
        reader = csv.reader(f)
        rows = []
        for row in reader:
            rows.append(row)
    return rows

def write(features, fileName, col_specs=None, title=None):
    with open(fileName, 'wb') as f:
        write_file(features,f,col_specs=col_specs, title=title)

def write_file(features, f, col_specs=None, title=None):
    
    # TODO auto gen col_specs from features
    assert col_specs

    writer = csv.writer(f)

    rows= []

    # title
    if title: rows.append([title])

    # columns
    cols = [spec[0] for spec in col_specs]
    rows.append(cols)

    # value rows
    for feature in features:
        props = feature['properties']
        row = [props[spec[2]] for spec in col_specs]
        rows.append(row)

    writer.writerows(rows)

def _test1(fname):
    col_specs=[('name',None,'FULLNAME'), ('CATEGORY',None,'CATEGORY')]

    features = [
    {
      "geometry": {
        "type": "LineString", 
        "coordinates": [ [10,11],[20,21] ]
        }, 
      "type": "Feature", 
      "properties": {
        "OBJECTID": 1, 
        "FULLNAME": "THE UPLANDS PATH", 
        "CATEGORY": "PEDESTRIAN"
      }
    },

    {
      "geometry": {
       "type": "LineString", 
        "coordinates": [ [100,110],[200,210] ]
        }, 
      "type": "Feature", 
      "properties": {
        "OBJECTID": 2, 
        "FULLNAME": "TRAIL", 
        "CATEGORY": "PEDESTRIAN"
      }
    },

    {
      "geometry": {
        "type": "Point", 
        "coordinates": [ 100,110 ]
        }, 
      "type": "Feature", 
      "properties": {
        "OBJECTID": 10, 
        "FULLNAME": "pointy dude",
        "CATEGORY": "LIGHT, RAIL STOP"
      }
    }]

    write(features,fname, col_specs=col_specs, title='Test Table')
    rows = read_table(fname)
    #print 'rows', rows
    assert len(rows) == 5
    assert rows[0][0] == 'Test Table'
    assert rows[1] == ['name','CATEGORY']
    assert rows[4] == ['pointy dude', 'LIGHT, RAIL STOP']

def test():
    with tempfile.NamedTemporaryFile(suffix='.csv', delete=False) as temp:
        fname = temp.name
    try:
        #print 'fname',fname
        _test1(fname)
    finally:
        if os.path.exists(fname): os.remove(fname)
    print "csvif PASS"

#doit
test()



'''
Implements simmple two dimensional bin system for objects 'indexed' by points: (x,y)
'''

import math

import geo

class Bins(object):
    def __init__(self,diameter):
        self.diameter = diameter
        self.bins = {}

    def _index(self,coord):
        return int(math.floor(coord/self.diameter))

    def _index2(self,point):
        x,y = point
        return (self._index(x), self._index(y))

    def add(self,point,value):
        bin = self.bins.setdefault(self._index2(point),[])
        bin.append((point,value))

    def in_rect(self,rect):
        x0,y0,x1,y1 = rect
        ix0 = self._index(x0)
        iy0 = self._index(y0)
        ix1 = self._index(x1)
        iy1 = self._index(y1)

        results = []
        for ix in range(ix0,ix1+1):
            for iy in range (iy0,iy1+1):
                if self.bins.has_key((ix,iy)):
                    for point,value in self.bins[(ix,iy)]:
                        if geo.point_in_rect_q(point, rect): results.append((point,value))
        return results

    def in_radius(self,point,r):
        px,py = point
        results = []
        for p2,value in self.in_rect((px-r,py-r,px+r,py+r)):
            if geo.distance(point,p2) < r: results.append((p2,value))
        return results
        
def test():
    b = Bins(100)
    b.add((1,1),'a')
    b.add((2,3),'b')
    b.add((1000,1000000),'far')
    assert len(b.in_rect((0,0,50,50))) == 2 
    assert len(b.in_rect((800,999000,1200,1001000))) == 1
    assert len(b.in_radius((1,1),0.5)) == 1
    assert len(b.in_radius((1,1),100)) == 2
    assert len(b.in_radius((1001,1000001),10)) == 1
    assert len(b.in_radius((-1,0),.1)) == 0

    print 'bins PASS'

#doit
test()




#!/usr/bin/env python
'''
Check that spiderosm is installed and functioning properly.
'''

import spiderosm.test
import spiderosm.config

#doit
assert __name__ == "__main__"
#get test data dir
spiderosm.config.read_config_files()

#run tests
spiderosm.test.run_tests()

#we only get here if all the tests pass.
print 'Congratulations!  SpiderOSM appears to be properly installed and functioning.'





#!/usr/bin/env python
'''
Match Berkeley street centerline to OSM
'''
import os
import pdb

import spiderosm.centerline
import spiderosm.config
import spiderosm.match

def _match_city():
    global spiderosm
    project = 'berkeley'

    # allow out_dir and gis_data_dir to be set in config files (.spiderosm.json)
    conf = spiderosm.config.settings
    conf['gis_data_dir'] = 'data'
    conf['out_dir'] = os.path.join('data',project)

    spiderosm.config.read_config_files()
    print 'spiderosm.config.settings:', spiderosm.config.settings

    gis_data_dir = conf['gis_data_dir']
    out_dir = conf['out_dir']
    print 'gis_data_dir=%s out_dir=%s' % (gis_data_dir,out_dir)

    # make sure out_dir exists
    if not os.path.exists(out_dir): os.makedirs(out_dir)

    # if a database is enabled write output to it too
    # NOTE: currently writes to only one database even if multiple databases are enabled.
    write_to_postgis = True
    write_to_spatialite = True
    db = None
    if write_to_postgis and (not db) and conf.get('postgis_enabled'):
        import spiderosm.postgis
        db = spiderosm.postgis.PGIS(project)
        print 'Results will be written to postgis database %s' % project
    if write_to_spatialite and (not db) and conf.get('spatialite_enabled'):
        import spiderosm.spatialite
        sqlite_fn = os.path.join(out_dir, project + '.sqlite')
        db = spiderosm.spatialite.Slite(sqlite_fn)
        print 'Results will be written to spatialite database %s' % sqlite_fn
       
    m = spiderosm.match.Match(
            project=project,
            proj4text='+proj=utm +zone=10 +ellps=WGS84 +units=m +no_defs',
            units='meters',
            db=db,
            out_dir=out_dir)

    #CITY DATA
    # source: http://www.ci.berkeley.ca.us/datacatalog/
    m.city_url = 'http://www.ci.berkeley.ca.us/uploadedFiles/IT/GIS/streets.zip'
    m.city_zip = os.path.join(gis_data_dir,'centerline','berkeley','streets','streets.zip')
    m.city_shp = os.path.join(gis_data_dir,'centerline','berkeley','streets','streets.shp')
    m.city_geojson = os.path.join(out_dir,'streets.geojson')
    m.centerline_to_pnwk = spiderosm.centerline.berkeley_pnwk
    m.city_network = os.path.join(out_dir,'city') # .pnwk.geojson

    #OSM DATA
    #geofabrik extracts updated daily
    m.osm_url = 'http://download.geofabrik.de/north-america/us/california-latest.osm.pbf'
    m.osm = os.path.join(gis_data_dir,'osm','geofabrik.de','california-latest.osm.pbf')
    m.osm_network = os.path.join(out_dir,'osm') # .pnwk.geojson

    #OSM BASE (before name fixes)
    #m.base = os.path.join(gis_data_dir,'osm','geofabrik.de','california-140401.osm.pbf')
    #m.base_network = os.path.join(out_dir,'base') # .pnwk.geojson

    # do the name crosscheck (results written to out_dir)
    m.names_cross_check()
    #m.names_osm_vs_base()

#doit
assert __name__ == "__main__"
_match_city()



#!/usr/bin/env python
'''
Match Portland Oregon RLIS street centerline to OSM

'''
import os

import spiderosm.centerline
import spiderosm.config
import spiderosm.geo
import spiderosm.match

def _match_city():
    global spiderosm
    project = 'portland'

    # allow out_dir and gis_data_dir and bbox to be set in config files (.spiderosm.json)
    # DATA DIRS
    conf = spiderosm.config.settings
    conf['gis_data_dir'] = 'data'
    conf['out_dir'] = os.path.join('data',project)

    #BOUNDING BOX
    #if bbox set to None, city data bounding box plus a buffer is used.
    # approximately portland city boundary, with 1 mile buffer.
    portland_bbox = (7604005.67,651319.12,7696978.42,732248.02)
    conf['bbox'] = spiderosm.geo.buffer_box(portland_bbox,5280) # buffer by 1 mi

    spiderosm.config.read_config_files()
    print 'spiderosm.config.settings:', spiderosm.config.settings

    gis_data_dir = conf['gis_data_dir']
    out_dir = conf['out_dir']
    bbox = conf['bbox']
    print 'gis_data_dir=%s out_dir=%s' % (gis_data_dir,out_dir)
    print 'bbox:', bbox 

    # make sure out_dir exists
    if not os.path.exists(out_dir): os.makedirs(out_dir)
    
    # if a database is enabled write output to it too
    # NOTE: currently writes to only one database even if multiple databases are enabled.
    write_to_postgis = True
    write_to_spatialite = True
    db = None
    if write_to_postgis and (not db) and conf.get('postgis_enabled'):
        import spiderosm.postgis
        db = spiderosm.postgis.PGIS(project)
        print 'Results will be written to postgis database %s' % project
    if write_to_spatialite and (not db) and conf.get('spatialite_enabled'):
        import spiderosm.spatialite
        sqlite_fn = os.path.join(out_dir, project + '.sqlite')
        db = spiderosm.spatialite.Slite(sqlite_fn)
        print 'Results will be written to spatialite database %s' % sqlite_fn

    m = spiderosm.match.Match(
            project=project,
            #NAD_1983_HARN_StatePlane_Oregon_North_FIPS_3601_Feet_Intl
            #http://spatialreference.org/ref/sr-org/6856/
            proj4text = '+proj=lcc +lat_1=44.33333333333334 +lat_2=46 +lat_0=43.66666666666666 +lon_0=-120.5 +x_0=2500000 +y_0=0 +datum=NAD83 +units=ft +no_defs',
            units='feet',
            bbox=bbox,
            db=db,
            out_dir=out_dir)

    # CITY DATA (RLIS street data)
    # source: http://rlisdiscovery.oregonmetro.gov/?resourceID=99
    # (No automatic download: go to site manually and download.)
    m.city_shp= os.path.join(gis_data_dir,'centerline','rlis','streets','streets.shp')
    m.city_geojson = os.path.join(out_dir,'streets.geojson')
    m.centerline_to_pnwk = spiderosm.centerline.rlis_pnwk
    m.city_network = os.path.join(out_dir,'city') # .pnwk.geojson

    #OSM DATA
    #geofabrik extracts updated daily
    m.osm_url = 'http://download.geofabrik.de/north-america/us/oregon-latest.osm.pbf'
    m.osm = os.path.join(gis_data_dir, 'osm', 'geofabrik.de', 'oregon-latest.osm.pbf')
    m.osm_network = os.path.join(out_dir,'osm') # .pnwk.geojson

    #OSM BASE (before name fixes)
    #m.base_url = 'http://download.geofabrik.de/north-america/us/oregon-140901.osm.pbf'
    #m.base = os.path.join(gis_data_dir,'osm','geofabrik.de','oregon-140901.osm.pbf')
    #m.base_network = os.path.join(out_dir,'base') # .pnwk.geojson

    # do the name crosscheck (results written to out_dir)
    m.names_cross_check()
    #m.names_osm_vs_base()

#doit
assert __name__ == "__main__"
_match_city()




'''
Utility class for using spiderosm to match jurisdictional (city) to (OSM) networks, generate name mismatch reports, etc.
'''

import cgi
import datetime
import json
import os
import pdb
import shutil
import tempfile

import geojson

import centerline
import config
import csvif
import geo
import geofeatures
import misc
import osm
import pnwk
import shp2geojson

class Match(object):
    def __init__(self, 
            project, 
            proj4text,
            units=None,
            bbox=None,
            out_dir=None,
            db = None):

        # defaults 
        if not out_dir: out_dir = 'data'
        if not units: units = 'meters'

        # properties
        self.project = project
        self.proj4text = proj4text
        self.units = units
        self.bbox = bbox
        self.out_dir = out_dir
        self.db = db;

        # city pnwk gen func
        self.city_centerline_to_pnwk = centerline.berkeley_pnwk        

        # city paths
        self.city_url = None
        self.city_zip = None
        self.city_shp = None
        self.city_geojson = None
        self.city_network = None # .pnwk.geojson

        # osm paths
        self.osm_url = None
        self.osm = None
        self.osm_network = None

        # base paths (osm base version)
        self.base_url = None
        self.base = None
        self.base_network = None

        self.log_current_task = None       
        # setup
        self.projection = geo.Projection(proj4text)
        if not os.path.exists(out_dir): os.makedirs(out_dir)

    def log(self,msg):
        now = str(datetime.datetime.now())
        task = self.log_current_task
        if not task: task=''

        print "=====",now,self.project,task,msg

    def log_begin_task(self,msg):
        assert not self.log_current_task
        self.log_current_task = msg
        self.log('...')

    def log_end_task(self):
        assert self.log_current_task
        self.log('DONE.')
        self.log_current_task = None
        
    def fetch_city_data(self):
        self.log_begin_task('fetching city data')
        if self.city_url:
            if self.city_zip:
                download_filename = self.city_zip
            else:
                download_filename = self.city_shp
            misc.update_file_from_url(download_filename, url=self.city_url)
        if self.city_zip:
            misc.unzip(self.city_zip)
        self.log_end_task()

    def fetch_osm_data(self):
        self.log_begin_task('fetching OSM data')
        if self.osm_url:
            misc.update_file_from_url(self.osm, url=self.osm_url)
        self.log_end_task()

    def fetch_base_data(self):
        self.log_begin_task('fetching OSM base data')
        if self.base_url:
            misc.update_file_from_url(self.base, url=self.base_url)
        self.log_end_task()

    def build_city_network(self):
        self.log_begin_task('building city network')
        # shp -> geojson
        if self.city_shp:
            shp2geojson.shp2geojson(self.city_shp, self.city_geojson, clip_rect=self.bbox)

        # geojson -> pnwk
        if self.city_geojson:
            with open(self.city_geojson) as f:
                city_geojson = geojson.load(f) 
            city_nwk = self.centerline_to_pnwk(name='city', features=city_geojson['features'], clip_rect=self.bbox)
            city_nwk.write_geojson(self.city_network) 
            if self.db: self.db.write_pnwk(city_nwk)
        self.log_end_task()

    def set_bbox_from_city(self):
        assert not self.bbox
        if not self.city_network: return
        self.log_begin_task('setting bbox from city network')
       
        city_nwk = pnwk.PNwk(name='city',filename=self.city_network,units=self.units)
        city_bbox = city_nwk.get_bbox()
        print 'city_bbox:', city_bbox

        # buffer
        if self.units == 'meters':
            city_bbox_buffered = geo.buffer_box(city_bbox,1000) #buffer by 1000 meters
        else:
            assert self.units == 'feet'
            city_bbox_buffered = geo.buffer_box(city_bbox,300) #buffer by 100 yards
        print 'bbox (city bbox with buffer):', city_bbox_buffered
        self.bbox = city_bbox_buffered

        #city_bbox_buffered_geo = conf['project_projection'].project_box(city_bbox_buffered, rev=True)
        #print 'city_bbox_buffered_geo:', city_bbox_buffered_geo
        self.log_end_task()

    def set_bbox_from_osm(self):
        assert not self.bbox
        if not self.osm_network: return
        self.log_begin_task('setting bbox from osm network')
       
        osm_nwk = pnwk.PNwk(name='osm',filename=self.osm_network,units=self.units)
        osm_bbox = osm_nwk.get_bbox()
        print 'osm_bbox:', osm_bbox

        # buffer
        if self.units == 'meters':
            osm_bbox_buffered = geo.buffer_box(osm_bbox,1000) #buffer by 1000 meters
        else:
            assert self.units == 'feet'
            osm_bbox_buffered = geo.buffer_box(osm_bbox,300) #buffer by 100 yards
        print 'bbox (osm bbox with buffer):', osm_bbox_buffered
        self.bbox = osm_bbox_buffered

        #city_bbox_buffered_geo = conf['project_projection'].project_box(city_bbox_buffered, rev=True)
        #print 'city_bbox_buffered_geo:', city_bbox_buffered_geo
        self.log_end_task()


    def build_osm_pnwk(self, name, in_fname, out_fname):
        if not in_fname: return

        # osm -> geojson
        osm_data = osm.OSMData(in_fname, clip_rect=self.bbox, target_proj=self.proj4text)
        osm_data.write_geojson(out_fname) # .osm.geojson
        if self.db: 
            self.db.write_geo(osm_data, name+'_ways',geometry_type='LineString')
            self.db.write_geo(osm_data, name+'_nodes',geometry_type='Point')

        # osm -> pnwk
        osm_nwk = osm_data.create_path_network(name=name)
        osm_nwk.write_geojson(out_fname) # .pnwk.geojson
        if self.db: 
            self.db.write_pnwk(osm_nwk,name+'_nwk')

    def build_osm_network(self):
        self.log_begin_task('building OSM network')
        self.build_osm_pnwk('osm', self.osm, self.osm_network)
        self.log_end_task()
    
    def build_base_network(self):
        self.log_begin_task('building OSM base network')
        self.build_osm_pnwk('base', self.base, self.base_network)
        self.log_end_task()
        
    def match_pnwks(self, pnwk1, pnwk2, match_suffix=None):
        if not match_suffix: match_suffix = '_matched'
        pnwk1_out = pnwk1.name + match_suffix
        pnwk2_out = pnwk2.name + match_suffix

        pnwk1.match(pnwk2)
        pnwk1.match_stats()
        pnwk2.match_stats()

        pnwk1.write_geojson(os.path.join(self.out_dir, pnwk1_out))
        pnwk2.write_geojson(os.path.join(self.out_dir, pnwk2_out))
        if self.db:
            self.db.write_pnwk(pnwk1, name=pnwk1_out)
            self.db.write_pnwk(pnwk2, name=pnwk2_out)

    def match_city_to_osm(self):
        if not self.city_network: return
        if not self.osm_network: return
        self.log_begin_task('matching city to osm network')
        city_nwk = pnwk.PNwk(filename=self.city_network, name='city',units=self.units)
        osm_nwk = pnwk.PNwk(filename=self.osm_network, name='osm',units=self.units)
        self.match_pnwks(osm_nwk, city_nwk)
        self.log_end_task()

    def match_osm_to_base(self):
        if not self.osm_network: return
        if not self.base_network: return
        self.log_begin_task('matching osm to base network')
        base_nwk = pnwk.PNwk(filename=self.base_network, name='base',units=self.units)
        osm_nwk = pnwk.PNwk(filename=self.osm_network, name='osm',units=self.units)
        self.match_pnwks(osm_nwk, base_nwk, match_suffix='_osm2base')
        self.log_end_task()

    def gen_mismatched_names_report(self):
        def mismatchFunc(feature,props):
            if props.get('match$score',0)<50: return False
            if props.get('match$score_name',100)==100: return False
            props['report$osm_verified'] = str(
                    ('osm$verified:name' in props) or 
                    ('osm$source:name' in props)) 
            props['report$wayURL'] = 'http://www.openstreetmap.org/way/%d' % props['osm$way_id']
            return True

        # read in matched network
        try:
            osm_matched = pnwk.PNwk(
                filename = os.path.join(self.out_dir, 'osm_matched'),
                name = 'osm',
                units = self.units)
        except IOError:
            self.log('Could not read osm_matched network, skipping mismatched_name_report')
            return

        self.log_begin_task('generating mismatched names report (.csv and .geojson)')

        # find mismatches
        mismatches = geofeatures.filter_features(osm_matched, 
                feature_func=mismatchFunc, 
                geom_type='LineString')
        print '%d (RAW) NAME MISMATCHES' % len(mismatches)

        # filter props down for webmap
        webmap_specs= [
                ('osm', 'TEXT', 'osm_pnwk$name'),
                ('city', 'TEXT', 'city_pnwk$name'),
                ('wayId', 'BIGINT', 'osm$way_id'),
                ('fixme', 'TEXT', 'osm$fixme:name'),
                ('osmVerified', 'TEXT', 'report$osm_verified')
                ]
        webmap = geofeatures.filter_features(mismatches, 
                col_specs=webmap_specs)

        # htlml quote TEXT property values to guard against injection attacks.
        for feature in webmap:
            props = feature.properties
            for (name, t, ignore) in webmap_specs:
                if t != 'TEXT': continue
                if not name in props: continue
                props[name] = cgi.escape(props[name])

        # unproject for web map use
        self.projection.project_geo_features(webmap,rev=True)
        
        # write out geojson for webmap
        geo = geojson.FeatureCollection(webmap)
        fname = os.path.join(self.out_dir,'name_mismatches.geojson')
        with open(fname, 'w') as f:
            geojson.dump(geo,f,indent=2,sort_keys=True)

        # make unique and sorted
        unique = []
        pairs = {}
        for feature in mismatches:
            props = feature['properties']
            pair = (props['osm_pnwk$name'], props['city_pnwk$name']) 
            if pair in pairs: continue
            unique.append(feature)
            pairs[pair] = True
        count = len(pairs)
        print '%d UNIQUE NAME MISMATCHES' % count
        def keyFunc(feature):
            p=feature['properties']
            return (p['osm_pnwk$name'], p['city_pnwk$name'])
        unique.sort(key=keyFunc)

        # write csv report
        title = '%s name mismatches between city centerline and OSM (generated: %s)'  
        title = title % (self.project, misc.date_ymdhms())

        csv_specs= [
                ('osm', 'TEXT', 'osm_pnwk$name'),
                ('city', 'TEXT', 'city_pnwk$name'),
                ('way', 'TEXT', 'report$wayURL'),
                ('osmVerified', 'TEXT', 'report$osm_verified')
                ]

        csvif.write(unique, os.path.join(self.out_dir,'name_mismatches.csv'), 
                col_specs=csv_specs,
                title=title)
        self.log_end_task()

    def gen_fixed_osm_names_report(self):
        def mismatchFunc(feature,props):
            if props.get('match$score',0)<50: return False
            if props.get('match$score_name',100)==100: return False
            if not props.get('osm$source:name'): return False
            props['report$wayURL'] = 'http://www.openstreetmap.org/way/%d' % props['osm$way_id']
            return True


        # read in matched network
        try:
            osm_matched = pnwk.PNwk(
                filename=os.path.join(self.out_dir,'osm_osm2base'),
                name = 'osm',
                units = self.units)
        except IOError:
            self.log('Could not read osm_osm2base matched network, skipping fixed (OSM) names report')
            return

        self.log_begin_task('generating fixed (OSM) names report (.csv and .geojson)')

        # find fixes
        fixes = geofeatures.filter_features(osm_matched, feature_func=mismatchFunc, geom_type='LineString')
        print '%d (RAW) NAME FIXES' % len(fixes)

        # filter props down for webmap
        webmap_specs= [
                ('old', 'TEXT', 'base$name'),
                ('new', 'TEXT', 'osm$name'),
                ('source','TEXT', 'osm$source:name'),
                ('wayId', 'BIGINT', 'osm$way_id')
                ]
        webmap = geofeatures.filter_features(fixes, col_specs=webmap_specs)

        # htlml quote TEXT property values to guard against injection attacks.
        for feature in webmap:
            props = feature.properties
            for (name, t, ignore) in webmap_specs:
                if t != 'TEXT': continue
                if not name in props: continue
                props[name] = cgi.escape(props[name])

        # unproject for web map use
        self.projection.project_geo_features(webmap,rev=True)
        
        # write out geojson for webmap
        geo = geojson.FeatureCollection(webmap)
        fname = os.path.join(self.out_dir,'name_fixes.geojson')
        with open(fname, 'w') as f:
            geojson.dump(geo,f,indent=2,sort_keys=True)

        # make unique and sorted
        unique = []
        pairs = {}
        for feature in fixes:
            props = feature['properties']
            pair = (props['osm_pnwk$name'], props['base_pnwk$name']) 
            if pair in pairs: continue
            unique.append(feature)
            pairs[pair] = True
        count = len(pairs)
        print '%d UNIQUE NAME FIXES' % count
        def keyFunc(feature):
            p=feature['properties']
            return (p['osm_pnwk$name'], p['base_pnwk$name'])
        unique.sort(key=keyFunc)
       
        # write csv report
        title = self.project + ' OSM name fixes (generated: %s)'  
        title = title % misc.date_ymdhms()

        csv_specs= [
                ('old', 'TEXT', 'base_pnwk$name'),
                ('new', 'TEXT', 'osm_pnwk$name'),
                ('way', 'TEXT', 'report$wayURL')
                ]

        csvif.write(unique, os.path.join(self.out_dir,'name_fixes.csv'), 
                col_specs=csv_specs,
                title=title)

        self.log_end_task()

    def names_cross_check(self):
        # build city network
        self.fetch_city_data()
        self.build_city_network()

        # if no bbox set, use (buffered) city bbox to clip OSM data
        if not self.bbox: self.set_bbox_from_city()

        # build OSM network
        self.fetch_osm_data()
        self.build_osm_network()

        # match networks
        self.match_city_to_osm()

        # gen mismatched names report (.csv and .geojson files)
        self.gen_mismatched_names_report()
        
    def names_osm_vs_base(self):
        # if no bbox set, use (buffered) city bbox to clip OSM data
        if not self.bbox: self.set_bbox_from_city()

        # build base network
        self.fetch_base_data()
        self.build_base_network()

        # match osm to base
        self.match_osm_to_base()

        # gen fixed osm names report (.csv and .geojson files)
        self.gen_fixed_osm_names_report()

def _test_ucb_sw1(out_dir):
    project = 'ucb_sw'
    test_data_dir = config.settings['spiderosm_test_data_dir']
    #print 'DEB test_data_dir:',test_data_dir
    in_dir = os.path.join(test_data_dir,'input',project)
    
    m = Match(
            project=project,
            proj4text='+proj=utm +zone=10 +ellps=WGS84 +units=m +no_defs',
            units='meters',
            out_dir=out_dir)

    #CITY DATA
    m.city_geojson = os.path.join(in_dir,'streets.geojson')
    m.centerline_to_pnwk = centerline.berkeley_pnwk
    m.city_network = os.path.join(out_dir,'city') # .pnwk.geojson

    #OSM DATA
    m.osm = os.path.join(in_dir,'ucb_sw.osm.xml')
    m.osm_network = os.path.join(out_dir,'osm') # .pnwk.geojson

    # do the name crosscheck (results written to out_dir)
    m.names_cross_check()

    # check results
    names_report_file = os.path.join(out_dir,'name_mismatches.csv')
    rows = csvif.read_table(names_report_file)
    assert len(rows) == 7
    assert rows[2] == ['FRANK SCHLESSINGER WAY',
            'CROSS CAMPUS RD',
            'http://www.openstreetmap.org/way/22278224',
            'True']

def test_ucb_sw():
    tmp_dir = tempfile.gettempdir()
    out_dir = os.path.join(tmp_dir,'ucb_sw')
    if os.path.exists(out_dir): shutil.rmtree(out_dir)
    try:
        _test_ucb_sw1(out_dir=out_dir)
    finally:
        if os.path.exists(out_dir): shutil.rmtree(out_dir)

def test():
    test_ucb_sw()
    print 'match PASS'
   
#doit
if __name__ == '__main__':
    config.read_config_files()
    test()



'''
Path Network - Matcher
'''
import pdb

import geo 
import pnwk_matchspokes

class PNwkMatch(pnwk_matchspokes.PNwkMatchSpokes):
        
    # match into nwk2 
    # matches are searched for in radius d and must be unique within that radius in both networks
    def match(self, nwk2, max_d=None, d=None, quiet=False):
        if not max_d:
            assert self.street_scale == nwk2.street_scale
            max_d = self.street_scale
        if not d:
            assert self.block_scale == nwk2.block_scale
            d = self.block_scale / 3.0

        #TODO remove or restore max_d,d

        # Check network consistency
        self.check()
        nwk2.check()

        # Initial stats
        statsBefore = nwk2.match_stats(quiet=True)

        # Identify corresponding junctions as launch points for spoke crawling.
        #
        # It's more important to be certain of matches, then to be comprehensive:
        # spoke crawling will fill in the gaps.
        self.jct_match_pass(nwk2, d, [(self.JctMatchSet.jct_filter_max_d, (max_d,d))])

        # Crawl out from matched juctions along 'spokes'
        # NOTE INITIAL jct matches are one <-> one.  But not true in general.
        new_jct_matches = []
        for jct in nwk2.jcts.values():
            if jct.match: new_jct_matches.append((jct.match, jct))
        #print 'DEBUG initial new_jct_matches:', new_jct_matches
        self.spoke_crawl(new_jct_matches, 
                msg='crawling from initial matched junctions',
                quiet=quiet)

        # try (re)crawling out from matched segment end points
        progress = True
        crawl = 0
        while progress:
            crawl += 1
            new_jct_matches = []
            for seg in self.segs.values():
                if not seg.match: continue
                if seg.match_rev:
                    new_jct_matches.append((seg.from_jct,seg.match.to_jct))
                    new_jct_matches.append((seg.to_jct,seg.match.from_jct))
                else:
                    new_jct_matches.append((seg.from_jct,seg.match.from_jct))
                    new_jct_matches.append((seg.to_jct,seg.match.to_jct))
            progress = self.spoke_crawl(new_jct_matches, 
                    msg='(re)craw number %d, from matched segment end points' % crawl,
                    quiet=quiet
                    )
            
        # Progress stats 
        statsAfter = nwk2.match_stats(quiet=True)
        num_segs_matched = statsAfter['num_segs_matched']
        delta = statsAfter['num_segs_matched'] - statsBefore['num_segs_matched']
        #if not quiet: 
        if False:
            print '===== new segs matches: %d, total seg matches: %d' % (
            delta, num_segs_matched)

        # Score matches
        self.score_matches()
        
        # share attributes of matched segs and jcts
        for seg in self.segs.values():
            if seg.match: 
                seg.tags.update(seg.match._asdict())
                seg.match.tags.update(seg._asdict())
        for jct in self.jcts.values():
            if jct.match: 
                jct.tags.update(jct.match.tags)
                jct.match.tags.update(seg.tags)

        # Check network consistency
        self.check()
        nwk2.check()

def test():
    points = [(0,0), (1000,1000)]
    points2 = [(5,0), (1005,1000)]
    g = PNwkMatch('g')    
    g.add_seg(1, 10, 11, points, names=['foo',
        "NE HALSEY ST FRONTAGE RD",
        "Northeast Halsey Street Frontage Road"])

    # test match
    n2 = PNwkMatch('n2_nwk')
    n2.add_seg(seg_id=100, from_jct_id=1000, to_jct_id=1100, points=points2, names=['bar'])
    g.match(n2, quiet=True)
    stats = g.match_stats(quiet=True)
    #stats = g.match_stats(quiet=False)
    assert stats['num_jcts_matched'] == 2 and stats['num_segs_matched'] == 1
    assert g.segs[1].match == n2.segs[100]   
    assert n2.segs[100].match_rev == 0
    g.check()
   
    print 'pnwk_match PASS'

#doit
test()




# bootstrap:  install setuptools, if needed
import ez_setup
ez_setup.use_setuptools()

from setuptools import setup

setup(name='spiderosm',
      version='0.2.0',
      description='GIS conflation tool for matching street networks.',
      long_description='GIS conflation tool: matches segments in one path network (e.g. streets and trails) to corresponding segments in another, based on geography and network connectivity.  Useful, among other things, for combining jurisdictional centerline data with Open Street Maps data.',
      classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Topic :: Scientific/Engineering :: GIS'
      ], 
      keywords='GIS conflation OSM OpenStreetMaps centerline jurisdictional street network matcher',
      url='http://spiderosm.org',
      author='Michael Arnold',
      author_email='mha@spiderosm.com',
      license='MIT',
      packages=['spiderosm'],
      install_requires=[
          'pyproj',
          'shapely', #requires geos library, on Mac: "%brew install geos"  
          'pylev',
          'geojson >= 1.0.9',
          # IMPOSM DOES NOT WORK UNDER WINDOWS!
          # imposm loads psycopg2 (postgis interface)
          # imposm needs protobuf / protoc, on Mac: "%brew install protobuf --with-python"
          # impost needs tokyo-cabinet, on Mac: "%brew install tokyo-cabinet"
          'imposm',  
          'pyshp' 
          ],
      extras_require={
          'spatialite' : ['pyspatialite >= 3.0.1']
          },
      include_package_data=True,
      scripts=[
          'spiderosm/bin/spiderosm_test.py',
          'spiderosm/bin/spiderosm_berkeley.py',
          'spiderosm/bin/spiderosm_portland.py'
      ],
      zip_safe=False)



'''
Network base for Path Network
'''

import json
import os
import pdb
import tempfile

import geojson

import bins
import cannames
import geo 
import pnwk_namespace

class PNwkNetwork(pnwk_namespace.PNwkNamespace):
    FILE_EXTENSION = '.pnwk.geojson'

    def __init__(self,name=None,filename=None, units=None, quiet=False):

        #network name
        if not name:
            if filename: 
                name = filename
            else:
                name = 'nwk'

        # parent init (namespaces)
        super(PNwkNetwork,self).__init__(name=name)

        # scale constants in meters
        self.step_scale_m  = 0.3    # approximately one step
        self.street_scale_m = 8.0   # typical street width
        self.block_scale_m = 100.0   # typical distance between intersections

        # units
        if not units: units="meters"
        if units=="meters":
            self.units_per_meter = 1.0
        else: 
            if units =="feet":
                self.units_per_meter = 3.28084
            else:
                print "Unrecognized units:", units
                assert False
        f = self.units_per_meter
        self.stepScale = self.step_scale_m * f
        self.street_scale = self.street_scale_m * f
        self.block_scale = self.block_scale_m * f
        
        # segs and jcts
        self.segs = {}
        self.jcts = {}
        self.jct_bins = bins.Bins(self.block_scale)  # bins for searching for jcts by location
        self.last_gen_seg_id = 0
        self.last_gen_jct_id = 0

        if filename:
            self.read_geojson(filename, quiet=quiet)
      
    # qualifiedTags = tags with namespace prefixes.
    def add_seg(self, seg_id, from_jct_id, to_jct_id, points, names=(), tags=None, qualifiedTags=None):
        assert not self.segs.has_key(seg_id)
        #assert from_jct_id != to_jct_id  # looping segs happen e.g. "CARRINGTON LN" in Portland.
        assert len(points)>=2

        if not seg_id: seg_id = self._gen_seg_id()

        # make names canonical to facilitate comparison between networks from different sources
        cans = set([])
        for name in names: 
            canName = cannames.canonical_street_name(name)
            if len(canName)>0: cans.add(cannames.canonical_street_name(name))

        tags = self.add_namespace(tags,self.client_name_space)
        if qualifiedTags: tags.update(qualifiedTags)

        seg = self.Seg(
                network=self, 
                seg_id=seg_id, 
                points=points, 
                names=cans, 
                tags=tags)
                
        seg.from_jct = self._add_connection(from_jct_id, seg, start=True)
        seg.to_jct = self._add_connection(to_jct_id, seg, start=False)
        self.segs[seg_id] = seg

    # explicit add jct not necessary - useful mainly for adding tags.
    # multiple adds of same jct_id ok.  point needs to match, tags are updated on each call
    # qualifiedTags = tags with namespace prefixes.
    def add_jct(self, jct_id, point, tags=None, qualifiedTags=None):
        if not self.jcts.has_key(jct_id): self._add_jct(jct_id, point)
        jct = self.jcts[jct_id]
        assert jct.point == point

        tags = self.add_namespace(tags,self.client_name_space)
        if qualifiedTags: tags.update(qualifiedTags)
        jct.tags.update(tags)

    def split_seg(self, seg, d, fromBeginning, new_jct_id=None):
        if not new_jct_id: new_jct_id = self._gen_jct_id()
        assert new_jct_id < 0
        assert not self.jcts.has_key(new_jct_id)

        if fromBeginning:
            points, points2 = geo.cut(seg.points, d)
            jct2 = seg.to_jct

            seg.points = points
            seg.to_jct.segs.remove(seg)
            seg.to_jct = self._add_connection(new_jct_id, seg, start=False)
            
            self.add_seg(
                    seg_id=None, 
                    from_jct_id=new_jct_id, 
                    to_jct_id=jct2.jct_id, 
                    points=points2, 
                    names=list(seg.names), 
                    qualifiedTags=seg.tags
                    )
        else: 
            points2, points = geo.cut(seg.points, seg.length()-d)
            jct2 = seg.from_jct

            seg.points = points
            seg.from_jct.segs.remove(seg)
            seg.from_jct = self._add_connection(new_jct_id, seg, start=True)
            self.add_seg(
                    seg_id=None, 
                    from_jct_id=jct2.jct_id, 
                    to_jct_id=new_jct_id, 
                    points=points2, 
                    names=list(seg.names), 
                    qualifiedTags=seg.tags
                    )

    # create a new jct_id
    # (generated ids are negative)
    def _gen_jct_id(self):
        self.last_gen_jct_id -= 1
        new = self.last_gen_jct_id
        assert new not in self.jcts
        return new 
   
    # create a new seg_id
    # (generated ids are negative)
    def _gen_seg_id(self):
        self.last_gen_seg_id -= 1
        new = self.last_gen_seg_id
        assert new not in self.segs
        return new 

    def _add_jct(self, jct_id, point):
        assert not self.jcts.has_key(jct_id)
        jct = self.Jct(network=self, jct_id=jct_id, point=point)
        self.jcts[jct_id] = jct 
        self.jct_bins.add(point, jct)

    def _add_connection(self, jct_id, seg, start):
        assert type(seg) == self.Seg
        # junction coords
        if (start):
            point=seg.points[0]
        else:
            point=seg.points[-1]

        if not self.jcts.has_key(jct_id): self._add_jct(jct_id, point)
        jct = self.jcts[jct_id]
        assert jct.point == point
        jct.segs.add(seg)
        
        return jct

    @property
    def __geo_interface__(self):
        features = []
        for seg in self.segs.values(): features.append(seg.__geo_interface__)
        for jct in self.jcts.values(): features.append(jct.__geo_interface__)
        return geojson.FeatureCollection(features)

    def _parse_geojson(self,geo):
        #print 'DEBUG _parse_geojson geo:', geo
        assert geo['type'] == 'FeatureCollection'

        # property namespace prefixes
        # TODO store pnwk name in json - for now name given at time of read MUST be the same.
        nsc = self.client_name_space
        nsp = self.pnwkNameSpace

        for feature in geo['features']:
            geometry = feature['geometry']
            properties = feature['properties']

            if geometry['type'] == 'Point':
                assert geometry['type'] == 'Point'
                jct_id = feature['id']
                point = tuple(geometry['coordinates'])

                # filter out pnwk properties
                tags={}
                for (key,value) in properties.items():
                    (ns, name) = self.split_off_namespace(key) 
                    if ns == nsp: continue
                    tags[key] = value
                #print 'tags',tags
                self.add_jct(jct_id, point, qualifiedTags=tags)

            else:
                assert geometry['type'] == 'LineString'
                seg_id = feature['id']
                points = [tuple(p) for p in geometry['coordinates']]
                from_jct_id = properties[nsp + 'from_jct_id']
                to_jct_id = properties[nsp + 'to_jct_id']
                points = [tuple(p) for p in geometry['coordinates']]
                try:
                    names = properties[nsp + 'name'].split(';')
                except KeyError:
                    names = ()

                # filter out pnwk properties
                tags={}
                for (key,value) in properties.items():
                    (ns, name) = self.split_off_namespace(key) 
                    if ns == nsp: continue
                    tags[key] = value

                self.add_seg(seg_id, from_jct_id, to_jct_id, points, names=names, qualifiedTags=tags)

    def read_geojson(self, filename, quiet=False): 
        if not quiet: print 'Reading', filename
        with open(filename+self.FILE_EXTENSION,'r') as f:
            # needs geojson>1.0.7, for non-ascii property keys
            geo = geojson.load(f)
        self._parse_geojson(geo)

    def write_geojson(self, name=None):
        if not name: name=self.name
        with open(name+self.FILE_EXTENSION,'w') as f:
            geojson.dump(self.__geo_interface__,f,indent=2,sort_keys=True)

    def get_bbox(self):
        bbox = geo.BBox() 
        for jct in self.jcts.values():
            bbox.add_point(jct.point)
        return bbox.rect()

    def get_jcts_near_point(self, point, max_d=100):
        result = []
        for point,jct in self.jct_bins.in_radius(point,max_d):
            result.append(jct)
        return result

    def meters(self,length):
        return length / self.units_per_meter

    def feet(self,length):
        return self.meters(length) * 3.28084

    def miles(self,length):
        return self.meters(length) * 0.000621371

    def check_jcts(self):
        for jct_id,jct in self.jcts.items():
            assert jct.jct_id == jct_id
            assert jct.network == self
            jct.check()

    def check_segs(self):
        for seg_id,seg in self.segs.items():
            assert seg.seg_id == seg_id
            assert seg.network == self
            seg.check()

    def check(self):
        self.check_jcts()
        self.check_segs()

    class Seg(object):
        def __init__(self, network, seg_id, points, from_jct=None, to_jct=None, names=None, tags=None):
            self.names = set([]) 
            self.tags = {}
            self.match = None

            self.network = network
            self.seg_id = seg_id
            self.points = points
            self.from_bearing=geo.compass_bearing(points[0],points[1])
            self.to_bearing=geo.compass_bearing(points[-1],points[-2])
            self.from_jct = from_jct
            self.to_jct = to_jct
            if names: self.names.update(names)
            if tags: self.tags.update(tags)

        @property
        def __geo_interface__(self):
            geometry = geojson.LineString(self.points)
            return geojson.Feature(geometry=geometry, id=self.seg_id, properties=self._asdict())

        def _asdict(self):
            d = {}
            d.update(self.tags)
            ns = self.network.pnwkNameSpace

            if self.match: 
                d[ns + 'match_id'] = self.match.seg_id
                d[ns + 'match_rev'] = self.match_rev
            d[ns + 'seg_id'] = self.seg_id
            d[ns + 'length'] = self.length()
            d[ns + 'from_bearing'] = self.from_bearing
            d[ns + 'to_bearing'] = self.to_bearing
            d[ns + 'from_jct_id'] = self.from_jct.jct_id
            d[ns + 'to_jct_id'] = self.to_jct.jct_id
            if len(self.names) > 0: 
                d[ns + 'name'] = self.names_text()
            return d

        def names_text(self):
            return ';'.join(list(self.names))

        def other_end(self,jct):
            if self.from_jct == jct: 
                return self.to_jct
            else:
                assert self.to_jct == jct
                return self.from_jct

        def length(self, units=None):
            l = geo.length(self.points)
            if not units: return l
            if units == "feet":   return self.network.feet(l)
            if units == "meters": return self.network.meters(l)
            if units == "miles":  return self.network.miles(l)
            assert False

        def split(self, d, fromBeginning, new_jct_id=None):
            self.network.split_seg(self, d, fromBeginning, new_jct_id)
        
        def check(self):
            assert self.network.segs[self.seg_id] == self
            assert self.from_jct.network == self.network
            assert self.from_jct.point == self.points[0]
            assert self.to_jct.network == self.network
            assert self.to_jct.point == self.points[-1]

    class Jct(object):
        def __init__(self, network, jct_id, point):
            self.network = network
            self.jct_id = jct_id
            self.point = point
            self.segs = set([])
            self.tags = {}
            self.match = None

        @property
        def __geo_interface__(self):
            geometry = geojson.Point(self.point)
            return geojson.Feature(geometry=geometry, id=self.jct_id, properties=self._asdict())

        def _asdict(self):
            d = {}
            d.update(self.tags)
            ns = self.network.pnwkNameSpace
            d[ns + 'jct_id'] = self.jct_id
            d[ns + 'seg_ids'] = [seg.seg_id for seg in self.segs]
            if self.match: d['ns + match_id'] = self.match.jct_id
            if 'import_src' in d:
                d[ns + 'import_src_id'] = d['import_src'].jct_id
                del d['import_src']
            if 'export_count' in d:
                d[ns + 'export_count'] = d['export_count']
                del d['export_count']
            return d

        def names(self):
            nameSet = set([])
            for seg in self.segs:
                nameSet = nameSet.union(seg.names)
            return nameSet

        def check(self):
            for seg in self.segs: 
                assert self.network.jcts[self.jct_id] == self
                assert seg.network == self.network
                assert seg.from_jct == self or seg.to_jct == self
                assert not seg.match or (seg == seg.match.match)

def test_id_gen():
    pn = PNwkNetwork('pn')
    pn.check()
    assert pn._gen_jct_id() == -1
    assert pn._gen_jct_id() == -2
    assert pn._gen_seg_id() == -1
    assert pn._gen_seg_id() == -2

def test_setup_g(units=None):
    g = PNwkNetwork('g',units=units)    
    g.add_seg(1, 10, 11, test_points, names=['foo',
        "NE HALSEY ST FRONTAGE RD",
        "Northeast Halsey Street Frontage Road"])

    tags = { 'ele':1000, 'signage':'some', u'osm$stamv\xe4g': u'yes' }
    g.add_jct(10, g.segs[1].points[0], tags=tags)

    g.check()
    return g

def test_add_seg():
    g = test_setup_g()

    #print g.segs
    #print g.jcts
    assert len(g.segs) == 1
    assert len(g.jcts) == 2
    seg = g.segs[1]
    assert len(seg.names) == 2
    assert seg.points[1] == test_points[1]
    assert seg.from_jct.jct_id == 10
    assert seg.to_jct.jct_id == 11

def test_add_jct():
    g = test_setup_g()
    #print 'jcts[10].tags', g.jcts[10].tags
    assert g.jcts[10].tags['g$signage'] == 'some'
    assert g.jcts[10].tags['g$ele'] == 1000

def test_geo_interface():
    g = test_setup_g()
    seg = g.segs[1]
    geo = seg.__geo_interface__
    #print 'geo:', geo
    assert geo['geometry']['type'] == 'LineString'
    assert geo['geometry']['coordinates'][1] == seg.points[1]
    assert geo['properties']['g_pnwk$from_jct_id'] == 10

def _test_read_write1(fname,fname2):
    g = test_setup_g()
    seg = g.segs[1]
     
    g.write_geojson(fname)
    g2 = PNwkNetwork(name='g', filename=fname, quiet=True)
    assert len(g2.segs) == 1
    assert len(g2.jcts) == 2
    seg = g2.segs[1]
    assert len(seg.names) == 2
    assert seg.points[1] == test_points[1]
    assert seg.from_jct.jct_id == 10
    assert seg.to_jct.jct_id == 11
    assert g2.jcts[10].tags['g$signage'] == 'some'
    assert g2.jcts[10].tags['g$ele'] == 1000
    g2.check()
    
    g2.write_geojson(fname2)
    g3 = PNwkNetwork(name='g', filename=fname2, quiet=True)
    assert len(g3.segs) == 1
    assert len(g3.jcts) == 2
    seg = g3.segs[1]
    assert len(seg.names) == 2
    assert seg.points[1] == test_points[1]
    assert seg.from_jct.jct_id == 10
    assert seg.to_jct.jct_id == 11
    assert g2.jcts[10].tags['g$signage'] == 'some'
    assert g2.jcts[10].tags['g$ele'] == 1000
    g3.check()

def test_read_write():
    tmp_dir = tempfile.gettempdir()
    fname = os.path.join(tmp_dir,'spiderosm_pnwk_network_test')
    fname2 = fname +'2'
    try:
        _test_read_write1(fname=fname,fname2=fname2)
    finally:
        if os.path.exists(fname): os.remove(fname)
        if os.path.exists(fname2): os.remove(fname2)

def test_get_jcts_near_point():
    g = test_setup_g()
    nj = g.get_jcts_near_point((0,0), max_d=1500)
    nj2 = g.get_jcts_near_point((0,0))
    #print nj
    #print 'nj2:',nj2
    assert len(nj) == 2
    assert len(nj2) == 1
    g.check()

def test_get_bbox():
    g = test_setup_g()
    box = g.get_bbox()
    #print 'box:',box
    assert box == (0,0,1000,1000)
    g.check()

def test_units():
    gft = test_setup_g(units="feet")
    gm = test_setup_g(units="meters")
    segft = gft.segs[1]
    segm = gm.segs[1]
    assert round(segft.length())== 1414 and round(segft.length(units="meters"))==431
    assert round(segm.length(units="feet"))==4640 and round(segm.length("miles")*100)==88

test_points = [(0,0), (1000,1000)]
test_points2 = [(5,0), (1005,1000)]
def test():
    test_units()
    test_id_gen()
    test_add_seg()
    test_add_jct()
    test_geo_interface()
    test_read_write()
    test_get_jcts_near_point()
    test_get_bbox()

    print 'pnwk_network PASS'

#doit
test()
 



'''
Path Network - Match Jcts.
'''
import pdb

import geo 
import pnwk_score

class PNwkMatchJcts(pnwk_score.PNwkScore):
    
    # match must be unique within both networks (for distance d)
    def jct_match_pass(self, otherNwk, d, filters):
        for jct in self.jcts.values():
                
            #skip previously merged junctions
            if jct.match: continue

            # must be unique match in own network, within d 
            self_match_set = self.JctMatchSet(set_nwk=self, jct=jct, aread=d, filter_list=filters)
            if len(self_match_set.set_jcts) !=1: continue
            assert self_match_set.set_jcts[0] == jct

            # must be unique match in other network, within d
            other_match_set = self.JctMatchSet(set_nwk=otherNwk, jct=jct, aread=d, filter_list=filters)
            if len(other_match_set.set_jcts) != 1: continue
            other_jct = other_match_set.set_jcts[0]

            # match!
            jct.match = other_jct
            other_jct.match = jct

    def match_stats(self, name=None, quiet=False):
        if not name: name = self.name
        num_segs = len(self.segs)
        num_jcts = len(self.jcts)
        miles = 0.0
        miles_matched = 0.0

        num_segs_matched = 0
        for seg in self.segs.values():
            miles += seg.length(units="miles")
            if seg.match != None: 
                num_segs_matched += 1
                miles_matched += seg.length(units="miles")
        percent_segs_matched= (100*num_segs_matched)/num_segs
        percent_miles_matched = round(100*miles_matched/miles)

        num_jcts_matched = 0
        for jct in self.jcts.values():
            if jct.match != None: num_jcts_matched += 1
        percent_jcts_matched= (100*num_jcts_matched)/num_jcts
        if not quiet:
            print '%s: jcts: %d/%d (%d%%) matched. segs: %d/%d (%d%%) matched.' % (
                name,
                num_jcts_matched, num_jcts, (100*num_jcts_matched)/num_jcts,
                num_segs_matched, num_segs, (100*num_segs_matched)/num_segs)
            print '%s: miles: %d/%d (%d%%) matched.' % (
                    name,
                    miles_matched,
                    miles,
                    percent_miles_matched)

        return {'num_jcts_matched':num_jcts_matched, 
                'num_jcts':num_jcts,
                'percent_jcts_matched':percent_jcts_matched,

                'num_segs_matched':num_segs_matched, 
                'num_segs':num_segs, 
                'percent_segs_matched':percent_segs_matched,

                'miles':miles,
                'miles_matched':miles_matched,
                'percent_miles_matched':percent_miles_matched}

    class JctMatchSet(object):
        def apply_filter(self, filter_func, parms):
            new_list = []
            for setJct in self.set_jcts:
                if filter_func(self, setJct, parms): new_list.append(setJct)
            self.set_jcts = new_list

        def apply_filters(self, filter_list):
            for filter_func,parms in filter_list:
                self.apply_filter(filter_func,parms)

        def __init__(self, set_nwk, jct, aread, filter_list):
            self.jct_nwk = jct.network
            self.jct = jct
            self.set_nwk = set_nwk
            self.set_jcts = set_nwk.get_jcts_near_point(jct.point, aread)
            self.apply_filters(filter_list)

        def jct_filter_two_names_in_common(self, setJct, parms):
            return len(setJct.names().intersection(self.jct.names())) >= 2

        # within max_d and no other point within d
        def jct_filter_max_d(self, setJct, parms):   
            (max_d,d) = parms

            #print 'DEBUG jct_filter_max_d max_d=%d' % max_d
            return (geo.distance(setJct.point,self.jct.point) <= max_d and
                    len(self.set_nwk.get_jcts_near_point(self.jct.point, d)) == 1)

def test_setup_g():
    g = PNwkMatchJcts('g')    
    g.add_seg(1, 10, 11, test_points, names=['foo',
        "NE HALSEY ST FRONTAGE RD",
        "Northeast Halsey Street Frontage Road"])
    g.check()
    return g

def test_setup_n():
    g = PNwkMatchJcts('n')    
    g.add_seg(1, 10, 11, test_points2, names=['foo',
        "NE HALSEY ST FRONTAGE RD",
        "Northeast Halsey Street Frontage Road"])
    g.check()
    return g

test_points = [(0,0), (1000,1000)]
test_points2 = [(5,0), (1020,1000)]
def test():
    g = test_setup_g()
    n = test_setup_n()
    g.jct_match_pass(n, 40, [(g.JctMatchSet.jct_filter_max_d, (10,100))])
    stats = g.match_stats(quiet=True)
    #print 'stats', stats
    assert stats['num_jcts_matched'] == 1
    print 'pnwk_matchjcts PASS'

#doit
test()




import os
import time
import shutil
import urllib
import math
import cmath

import pyproj
import imposm.parser  # parses OpenStreetMaps xml and pbf file formats.
import geojson

import geo
import pnwk

osm_geojson_file_extension = '.osm.geojson'

class Way(object):
    def __init__(self, tags, node_ids):
        self.tags = tags
        self.node_ids = node_ids

class Node(object):
    def __init__(self, coords, tags=None):
        self.coords = coords
        self.tags = tags or {}       
        self.way_ids = {} # keys are way_ids of ways referencing this node.  values are ref counts.

class OSMData(object):
    ways = {}
    nodes = {}
    #relations = {} CURRENTLY NOT PARSING RELATIONS.

    def _parse_node(self, osmId, coords, tags=None):
        if self.proj: coords = self.proj.project_point(coords)
        if self.clip_rect and not geo.point_in_rect_q(coords,self.clip_rect): return
	self.nodes[osmId] = Node(coords, tags=tags)
        #if tags: print 'DEB _parse_node, osmId:',osmId,'tags:',tags

    # callback for coords (called for ALL nodes, both those with and without tags.)
    def _parse_coords(self, coords_in):
        for osmId, lon, lat in coords_in:
	    if not (osmId in self.nodes): self._parse_node(osmId, (lon,lat))

    # callback for nodes (with tags only)
    def _parse_nodes(self, nodes_in):
	for osmId,tags,coords in nodes_in:
            self._parse_node(osmId, coords, tags=tags)
       
    # callback for ways
    def _parse_ways(self, ways_in):
        for osmid, tags, refs in ways_in:

	    # roads only
	    if not ('highway' in tags): continue
            
            # at least two nodes (in clip area)
            count = 0
            for ref in refs:
                if ref in self.nodes: 
                    count += 1
                    if count >= 2: break
            if count < 2: continue

            # remove refs to missing nodes
            new = []
            for ref in refs:
                if ref in self.nodes: new.append(ref)
            refs = new

	    self.ways[osmid] = Way(tags=tags, node_ids=refs)		
 
    # read in an OSM data file
    # processing in two passes to avoid thrashing when clipping an area from huge files.
    def _parse_input_file(self, file_name):

        # first pass: nodes 
	p = imposm.parser.OSMParser(
	    coords_callback=self._parse_coords,
	    nodes_callback=self._parse_nodes)
				
	print 'Reading nodes from osm file:', file_name
	if os.path.splitext(file_name)[1] == '.xml':
	    p.parse_xml_file(file_name)
	else:
	    p.parse(file_name)

        # second pass: ways
	p = imposm.parser.OSMParser(
	    ways_callback=self._parse_ways)
				
	print 'Reading ways from osm file:', file_name
	if os.path.splitext(file_name)[1] == '.xml':
	    p.parse_xml_file(file_name)
	else:
	    p.parse(file_name)
	    
	print 'len(ways):', len(self.ways)
	print 'len(nodes):', len(self.nodes)

    # remove stutters (repeated node refs) from ways
    def _remove_stutters(self):
        num_removed = 0
        for way_id in self.ways:
            new = []
            for ref in self.ways[way_id].node_ids:
                if len(new) > 0 and ref == new[-1]:
                    print 'DELETING REPEATED NODE REFERENCE %d from way %d' % (ref, way_id)
                    num_removed += 1
                else:
                    new.append(ref)
            self.ways[way_id].node_ids = new 
        if num_removed>0: print 'REMOVED %d REPEATED NODE REFERENCES.' % num_removed

    # remove ways with less than two nodes left (after stutter removal.)
    def _remove_malformed_ways(self):
        num_removed = 0
        for way_id in self.ways.keys():
            if len(self.ways[way_id].node_ids) < 2: del self.ways[way_id]

    # add referencing ways to nodes
    def _add_refs_to_nodes(self):
	for way_id,way in self.ways.items():
	    for node_id in way.node_ids:
		node = self.nodes[node_id]
                # node.way_ids keys are ids for referencing Ways, value is ref count for that way.
		if way_id in node.way_ids: 
		    node.way_ids[way_id] += 1
		else: 
		    node.way_ids[way_id] = 1		
                    
    # delete nodes that have no tags and no refs
    def _vacuum_nodes(self):
	for node_id,node in self.nodes.items():
	    if len(node.tags)==0 and len(node.way_ids)==0: 
	        del self.nodes[node_id]
		#print 'del node:',node_id

    # initialize from OSM input file
    def __init__(self, file_name, clip_rect=None, target_proj=None):
        self.clip_rect = clip_rect

        self.proj=None
        if target_proj: self.proj = geo.Projection(target_proj)
       
	#read in file
	self._parse_input_file(file_name)

        #remove stutters from ways (repeated node refs)
        self._remove_stutters()

        # remove any ways with less than two nodes left.
        self._remove_malformed_ways()

	# add way references to nodes
	self._add_refs_to_nodes()

	# remove nodes that have neither tags nor are referenced by ways
	self._vacuum_nodes()
	print 'len(nodes) after vacuum:', len(self.nodes)

	# count intersections
	numi = 0
	for id,node in self.nodes.items():
	    if len(node.way_ids)>1: numi += 1
	print 'intersections:', numi

    def node_geo(self,node_id):
        node = self.nodes[node_id]
        geometry = geojson.Point(node.coords)
        properties = { 'nodeId' : node_id }
        properties.update(node.tags)
        return geojson.Feature(geometry=geometry, id=node_id, properties=properties)

    def way_geo(self, way_id):
        way = self.ways[way_id]
        points = [self.nodes[node_id].coords for node_id in way.node_ids]
        geometry = geojson.LineString(points)
        properties = {'way_id' : way_id}
        properties.update(way.tags)
        return geojson.Feature(geometry=geometry, id=way_id, properties=properties)
    
    @property
    def __geo_interface__(self):
        features = []
        for way_id in self.ways.keys():
            features.append(self.way_geo(way_id))
        for node_id in self.nodes.keys():
            features.append(self.node_geo(node_id))
        return geojson.FeatureCollection(features)

    def write_geojson(self, name):
        with open(name+osm_geojson_file_extension,'w') as f:
            geojson.dump(self.__geo_interface__,f,indent=2)
        
    # segments split at junctions
    # all attributes (keys) copied unless props/node_props specified.
    def create_path_network(self, name=None, seg_props=None, jct_props=None):
        if not name: name='osm'
        nwk = pnwk.PNwk(name=name)
        num_segs=0

        def add_node_tags(nwk, node_id, point, jct_props):
            node = self.nodes[node_id]
            tags = {'node_id': node_id}
            if not jct_props:
                tags.update(node.tags)
            else:
                for prop in jct_props:
                    if prop in node.tags: 
                        tags[prop] = node.tags[prop]

            nwk.add_jct(node_id, point, tags=tags)

        # process in sorted order, so that result is the same from run to run, despite multi-core processing
        way_ids = self.ways.keys()
        way_ids.sort()
        #print 'DEB way_ids:', way_ids
	for way_id in way_ids:
                way = self.ways[way_id]
		snode_id = way.node_ids[0]
                names = []
                if way.tags.has_key('name'): names.append(way.tags['name'])
                if way.tags.has_key('alt_name'): names.append(way.tags['alt_name'])
                if way.tags.has_key('loc_name'): names.append(way.tags['loc_name'])
                if way.tags.has_key('old_name'): names.append(way.tags['old_name'])
                if way.tags.has_key('official_name'): names.append(way.tags['official_name'])
                if way.tags.has_key('ref'): names += way.tags['ref'].split(';')

		points = []
                num = len(way.node_ids)
		last = num-1
		for i in xrange(num):
			node_id = way.node_ids[i]
			node = self.nodes[node_id]
			assert len(node.way_ids)>0
			points.append(node.coords)
			if len(points)<2 or (i!=last and len(node.way_ids)<2): continue
                        
                        #tags
                        tags = {'way_id':way_id}
                        if not seg_props:
                            tags.update(way.tags)
                        else:
                            for prop in seg_props:
                                if prop in way.tags: 
                                    tags[prop] = way.tags[prop]

                        num_segs += 1  # segment number used as id
                        nwk.add_seg(num_segs, snode_id, node_id, points, names=names, tags=tags)
                        add_node_tags(nwk, snode_id, points[0], jct_props)
                        add_node_tags(nwk, node_id, points[-1], jct_props)

                        snode_id = node_id;
                        points = [node.coords]
        print 'segments added:', num_segs
        return nwk

def test():
    # TODO: more rigorous testing.
    print "osm PASS"

#doit
test()

